"""
Tests to ensure we reliably wrap the Anthropic API.
"""

import asyncio
import os
import time
import unittest
from typing import Any, Dict

import anthropic
import pytest

from braintrust import logger
from braintrust.logger import ObjectMetadata, OrgProjectMetadata
from braintrust.util import LazyValue
from braintrust.wrappers.anthropic import wrap_anthropic

TEST_ORG_ID = "test-org-123"
PROJECT_NAME = "test-anthropic-app"
MODEL = "claude-3-haiku-20240307"  # use the cheapest model since answers dont matter


def _setup_test_logger(project_name: str):
    # FIXME[matt] make reusable
    project_metadata = ObjectMetadata(id=project_name, name=project_name, full_info=dict())
    metadata = OrgProjectMetadata(org_id=TEST_ORG_ID, project=project_metadata)
    lazy_metadata = LazyValue(lambda: metadata, use_mutex=False)
    l = logger.init_logger(project=project_name)
    l._lazy_metadata = lazy_metadata  # FIXME[matt] this is cheesy but it stops us from having to login


def _get_client():
    return anthropic.Anthropic()


def _get_async_client():
    return anthropic.AsyncAnthropic()


def test_memory_logger():
    # FIXME[matt] this should be moved to a common place
    _setup_test_logger("test-anthropic-app")
    with logger._internal_with_memory_background_logger() as bgl:
        assert not bgl.pop()

        @logger.traced
        def thing():
            return "hello"

        thing()
        logs = bgl.pop()
        assert len(logs) == 1
        assert logs


@pytest.fixture
def memory_logger():
    _setup_test_logger(PROJECT_NAME)
    with logger._internal_with_memory_background_logger() as bgl:
        yield bgl


def test_anthropic_messages_create_stream_true(memory_logger):
    assert not memory_logger.pop()

    client = wrap_anthropic(_get_client())
    kws = {
        "model": MODEL,
        "max_tokens": 300,
        "messages": [{"role": "user", "content": "what's the capital of Canada?"}],
        "stream": True,
    }

    start = time.time()
    with client.messages.create(**kws) as out:
        msgs = [m for m in out]
    end = time.time()

    assert msgs  # a very coarse grained check that this works

    logs = memory_logger.pop()
    assert len(logs) == 1
    log = logs[0][0]
    assert log["metadata"]["model"] == MODEL
    assert log["metadata"]["max_tokens"] == 300
    assert start < log["metrics"]["start"] < log["metrics"]["end"] < end


def test_anthropic_messages_model_params_inputs(memory_logger):
    assert not memory_logger.pop()
    client = wrap_anthropic(_get_client())

    kw = {
        "model": MODEL,
        "max_tokens": 300,
        "system": "just return the number",
        "messages": [{"role": "user", "content": "what is 1+1?"}],
        "temperature": 0.5,
        "top_p": 0.5,
    }

    def _with_messages_create():
        return client.messages.create(**kw)

    def _with_messages_stream():
        with client.messages.stream(**kw) as stream:
            for msg in stream:
                pass
        return stream.get_final_message()

    for f in [_with_messages_create, _with_messages_stream]:
        msg = f()
        assert msg.content[0].text == "2"

        logs = memory_logger.pop()
        assert len(logs) == 1
        log = logs[0][0]
        import pprint

        pprint.pprint(log)
        assert log["metadata"]["model"] == MODEL
        assert log["metadata"]["max_tokens"] == 300
        assert log["metadata"]["temperature"] == 0.5
        assert log["metadata"]["top_p"] == 0.5


def test_anthropic_messages_system_prompt_inputs(memory_logger):
    assert not memory_logger.pop()

    client = wrap_anthropic(_get_client())
    system = "Today's date is 2024-03-26. Only return the date"
    q = [{"role": "user", "content": "what is tomorrow's date? only return the date"}]

    args = {
        "messages": q,
        "temperature": 0,
        "max_tokens": 300,
        "system": system,
        "model": MODEL,
    }

    def _with_messages_create():
        return client.messages.create(**args)

    def _with_messages_stream():
        with client.messages.stream(**args) as stream:
            for msg in stream:
                pass
        return stream.get_final_message()

    for f in [_with_messages_create, _with_messages_stream]:
        print("testing %s" % f.__name__)
        msg = f()
        assert "2024-03-27" in msg.content[0].text

        logs = memory_logger.pop()
        assert len(logs) == 1
        log = logs[0][0]
        inputs = log["input"]
        assert len(inputs) == 2
        inputs_by_role = {m["role"]: m["content"] for m in inputs}
        assert inputs_by_role["system"] == system
        assert inputs_by_role["user"] == q[0]["content"]


@pytest.mark.asyncio
async def test_anthropic_messages_streaming_async(memory_logger):
    assert not memory_logger.pop()

    client = wrap_anthropic(_get_async_client())
    msgs_in = [{"role": "user", "content": "what is 1+1?, just return the number"}]

    start = time.time()
    async with client.messages.stream(max_tokens=1024, messages=msgs_in, model=MODEL) as stream:
        async for event in stream:
            pass
        msg_out = await stream.get_final_message()
        end = time.time()
        assert msg_out.content[0].text == "2"
        usage = msg_out.usage

        logs = memory_logger.pop()
        assert len(logs) == 1
        log = logs[0][0]
        assert "user" in str(log["input"])
        assert "1+1" in str(log["input"])
        assert "2" in str(log["output"])
        assert log["project_id"] == PROJECT_NAME
        assert log["span_attributes"]["type"] == "llm"
        assert log["metadata"]["model"] == MODEL
        assert log["metadata"]["max_tokens"] == 1024
        _assert_metrics_are_valid(log["metrics"])
        assert start < log["metrics"]["start"] < log["metrics"]["end"] < end
        metrics = log["metrics"]
        assert metrics["prompt_tokens"] == usage.input_tokens
        assert metrics["completion_tokens"] == usage.output_tokens
        assert metrics["tokens"] == usage.input_tokens + usage.output_tokens
        assert metrics["cache_read_input_tokens"] == usage.cache_read_input_tokens
        assert metrics["cache_creation_input_tokens"] == usage.cache_creation_input_tokens
        assert log["metadata"]["model"] == MODEL
        assert log["metadata"]["max_tokens"] == 1024


def test_anthropic_client_error(memory_logger):
    assert not memory_logger.pop()

    client = wrap_anthropic(_get_client())

    fake_model = "there-is-no-such-model"
    msg_in = {"role": "user", "content": "who are you?"}

    try:
        client.messages.create(model=fake_model, max_tokens=999, messages=[msg_in])
    except Exception:
        pass
    else:
        raise Exception("should have raised an exception")

    logs = memory_logger.pop()
    assert len(logs) == 1
    log = logs[0][0]
    assert log["project_id"] == PROJECT_NAME
    assert "404" in log["error"]


def test_anthropic_messages_streaming_sync(memory_logger):
    assert not memory_logger.pop()

    client = wrap_anthropic(_get_client())
    msg_in = {"role": "user", "content": "what is 2+2? (just the number)"}

    start = time.time()
    with client.messages.stream(model=MODEL, max_tokens=300, messages=[msg_in]) as stream:
        msgs_out = [m for m in stream]
    end = time.time()
    msg_out = stream.get_final_message()
    usage = msg_out.usage
    # crudely check that the stream is valid
    assert len(msgs_out) > 3
    assert 1 <= len([m for m in msgs_out if m.type == "text"])
    assert msgs_out[0].type == "message_start"
    assert msgs_out[-1].type == "message_stop"

    logs = memory_logger.pop()
    assert len(logs) == 1
    log = logs[0][0]
    assert "user" in str(log["input"])
    assert "2+2" in str(log["input"])
    assert "4" in str(log["output"])
    assert log["project_id"] == PROJECT_NAME
    assert start < log["metrics"]["start"] < log["metrics"]["end"] < end
    assert log["span_attributes"]["type"] == "llm"
    _assert_metrics_are_valid(log["metrics"])
    assert log["metrics"]["prompt_tokens"] == usage.input_tokens
    assert log["metrics"]["completion_tokens"] == usage.output_tokens
    assert log["metrics"]["tokens"] == usage.input_tokens + usage.output_tokens
    assert log["metrics"]["cache_read_input_tokens"] == usage.cache_read_input_tokens
    assert log["metrics"]["cache_creation_input_tokens"] == usage.cache_creation_input_tokens


def test_anthropic_messages_sync(memory_logger):
    assert not memory_logger.pop()

    client = wrap_anthropic(_get_client())

    msg_in = {"role": "user", "content": "what's 2+2?"}

    start = time.time()
    msg = client.messages.create(model=MODEL, max_tokens=300, messages=[msg_in])
    end = time.time()

    text = msg.content[0].text
    assert text

    # verify we generated the right spans.
    logs = memory_logger.pop()

    assert len(logs) == 1
    log = logs[0][0]
    assert "2+2" in str(log["input"])
    assert "4" in str(log["output"])
    assert log["project_id"] == PROJECT_NAME
    assert start < log["metrics"]["start"] < end
    assert start < log["metrics"]["end"] < end
    assert log["span_id"]
    assert log["root_span_id"]
    attrs = log["span_attributes"]
    assert attrs["type"] == "llm"
    assert "anthropic" in attrs["name"]
    metrics = log["metrics"]
    _assert_metrics_are_valid(metrics)
    assert start < metrics["start"] < metrics["end"] < end
    assert log["metadata"]["model"] == MODEL


def _assert_metrics_are_valid(metrics: Dict[str, Any]):
    assert metrics["tokens"] > 0
    assert metrics["prompt_tokens"] > 0
    assert metrics["completion_tokens"] > 0



"""
A module providing an LRU (Least Recently Used) cache implementation.

This module contains a generic LRU cache that can store key-value pairs of any type.
The cache maintains items in order of use and can optionally evict least recently
used items when it reaches a maximum size. The implementation uses an OrderedDict
for O(1) access and update operations.
"""

from typing import Generic, Optional, OrderedDict, TypeVar

K = TypeVar("K")
V = TypeVar("V")


class LRUCache(Generic[K, V]):
    """
    A Least Recently Used (LRU) cache implementation.

    This cache maintains items in order of use, evicting the least recently used item
    when the cache reaches its maximum size (if specified). Items are considered "used"
    when they are either added to the cache or retrieved from it.

    If no maximum size is specified, the cache will grow unbounded.

    Args:
        max_size: Maximum number of items to store in the cache.
                 If not specified, the cache will grow unbounded.
    """

    def __init__(self, max_size: Optional[int] = None):
        self._cache: OrderedDict[K, V] = OrderedDict()
        self._max_size = max_size

    def get(self, key: K) -> V:
        """
        Retrieves a value from the cache.
        If the key exists, the item is marked as most recently used.

        Args:
            key: The key to look up.

        Returns:
            The cached value.

        Raises:
            KeyError: If the key is not found in the cache.
        """
        if key not in self._cache:
            raise KeyError(f"Cache key not found: {key}")

        # Refresh key by moving to end of OrderedDict.
        value = self._cache.pop(key)
        self._cache[key] = value
        return value

    def set(self, key: K, value: V) -> None:
        """
        Stores a value in the cache.
        If the key already exists, the value is updated and marked as most recently used.
        If the cache is at its maximum size, the least recently used item is evicted.

        Args:
            key: The key to store.
            value: The value to store.
        """
        if key in self._cache:
            self._cache.pop(key)
        elif self._max_size and len(self._cache) >= self._max_size:
            # Remove oldest item (first item in ordered dict).
            self._cache.popitem(last=False)

        self._cache[key] = value

    def clear(self) -> None:
        """Removes all items from the cache."""
        self._cache.clear()



"""Implements the braintrust push subcommand."""

import importlib.abc
import importlib.machinery
import importlib.metadata
import importlib.util
import inspect
import json
import os
import re
import subprocess
import sys
import tempfile
import textwrap
import zipfile
from typing import Any, Dict, List, Optional

import requests

from .. import api_conn, app_conn, login, org_id, proxy_conn
from ..framework2 import CodeFunction, global_
from ..types import IfExists
from ..util import add_azure_blob_headers


class _ProjectIdCache:
    def __init__(self):
        self._cache = {}

    def get(self, project):
        if project not in self._cache:
            resp = app_conn().post_json("api/project/register", {"project_name": project.name})
            self._cache[project] = resp["project"]["id"]
        return self._cache[project]


def _pkg_install_arg(pkg) -> Optional[str]:
    try:
        dist = importlib.metadata.distribution(pkg)
        direct_url = dist._path / "direct_url.json"  # type: ignore
        if direct_url.exists():
            with open(direct_url) as f:
                j = json.loads(f.read())
                if "url" in j:
                    return j["url"]
        return f"{pkg}=={dist.version}"
    except importlib.metadata.PackageNotFoundError as e:
        print(f"Failed to find package {pkg}: {e}", file=sys.stderr)
    return None


def _pydantic_to_json_schema(m):
    if hasattr(m, "model_json_schema"):
        # pydantic 2
        return m.model_json_schema()
    # pydantic 1
    return m.schema()


def _check_uv():
    try:
        import uv as _
    except ImportError:
        raise ValueError(
            textwrap.dedent(
                f"""\
                The `uv` package is required to push to Braintrust. You can install it by including the
                extra "cli" dependencies. Run:

                  pip install 'braintrust[cli]'

                to install braintrust with the CLI dependencies (make sure to quote 'braintrust[cli]')."""
            ),
        )


class _ProjectRootImporter(importlib.abc.MetaPathFinder):
    """An importer that only resolves top-level modules from the project root and their submodules,
    and collects the source files of all imported modules.
    """

    def __init__(self) -> None:
        self._project_root, self._path_rest = sys.path[0], sys.path[1:]
        self._sources = []

    def _under_project_root(self, path: List[str]) -> bool:
        """Returns true if all paths in `path` are under the project root."""
        return all(p.startswith(self._project_root) for p in path)

    def _under_rest(self, path: List[str]) -> bool:
        """Returns true if any path in `path` is under one of the remaining paths in `sys.path`."""
        return any(p.startswith(pr) for p in path for pr in self._path_rest)

    def find_spec(self, fullname, path, target=None):
        if path is None:
            # Resolve top-level modules only from the project root.
            path = [self._project_root]
        elif not self._under_project_root(path) or self._under_rest(path):
            # Defer paths that are not under the project root or covered by another sys.path entry
            # to the subsequent importers.
            return None
        spec = importlib.machinery.PathFinder.find_spec(fullname, path, target)
        if spec is not None and spec.origin is not None:
            self._sources.append(spec.origin)
        return spec

    def sources(self) -> List[str]:
        return self._sources


def _import_module(name: str, path: str) -> List[str]:
    """Imports the module and returns the list of source files
    of all modules imported in the process.

    Args:
        name: The fully qualified name of the module to import.
        path: The absolute path to the module to import.

    Returns:
        A list of absolute paths to source files of all modules imported in the process.
    """
    importer = _ProjectRootImporter()
    sys.meta_path.insert(0, importer)

    importlib.import_module(name)
    return importer.sources()


def _py_version() -> str:
    return f"{sys.version_info.major}.{sys.version_info.minor}"


def _run_install(install_args: List[str], packages_dir: str):
    subprocess.run(
        [
            "uv",
            "pip",
            "install",
            *install_args,
            "--target",
            packages_dir,
            "--python-platform",
            os.environ.get("BRAINTRUST_INTERNAL_PY_BUNDLE_PLATFORM_OVERRIDE", "linux"),
            "--python-version",
            os.environ.get("BRAINTRUST_INTERNAL_PY_BUNDLE_VERSION_OVERRIDE", _py_version()),
        ],
        check=True,
    )


def _upload_bundle(entry_module_name: str, sources: List[str], requirements: Optional[str]) -> str:
    _check_uv()

    resp = proxy_conn().post_json(
        "function/code",
        {
            "org_id": org_id(),
            "runtime_context": {
                "runtime": "python",
                "version": _py_version(),
            },
        },
    )
    bundle_upload_url = resp["url"]
    bundle_id = resp["bundleId"]

    with tempfile.TemporaryDirectory() as td:
        packages_dir = os.path.join(td, "pkg")

        # Though not strictly necessary, these packages should be those supported in the Python code editor
        # with the exception of pydantic, which is necessary to allow the user to express function input schemas.
        _run_install(
            [
                arg
                for arg in [
                    _pkg_install_arg(pkg)
                    for pkg in [
                        "pydantic",
                        "braintrust",
                        "autoevals",
                        "requests",
                        "openai",
                    ]
                ]
                if arg is not None
            ],
            packages_dir,
        )
        if requirements:
            # Overwrite any packages that are already installed.
            _run_install(["--requirement", requirements], packages_dir)

        with zipfile.ZipFile(
            os.path.join(td, "pkg.zip"), "w", compression=zipfile.ZIP_DEFLATED, compresslevel=9
        ) as zf:
            for dirpath, dirnames, filenames in os.walk(packages_dir):
                arcdirpath = os.path.relpath(dirpath, packages_dir)
                arcdirpath = os.path.normpath(arcdirpath)
                for name in sorted(dirnames):
                    path = os.path.join(dirpath, name)
                    arcname = os.path.join(arcdirpath, name)
                    zf.write(path, arcname)
                for name in filenames:
                    path = os.path.join(dirpath, name)
                    path = os.path.normpath(path)
                    if os.path.isfile(path):
                        arcname = os.path.join(arcdirpath, name)
                        zf.write(path, arcname)
            for source in sources:
                zf.write(source, os.path.relpath(source))
            zf.writestr("register.py", f"import {entry_module_name} as _\n")
        headers = {}
        add_azure_blob_headers(headers, bundle_upload_url)
        with open(os.path.join(td, "pkg.zip"), "rb") as zf:
            requests.put(
                bundle_upload_url,
                data=zf.read(),
                headers=headers,
            ).raise_for_status()

    return bundle_id


def _collect_function_function_defs(
    project_ids: _ProjectIdCache, functions: List[Dict[str, Any]], bundle_id: str, if_exists: IfExists
) -> None:
    for i, f in enumerate(global_.functions):
        source = inspect.getsource(f.handler)
        if f.handler.__name__ == "<lambda>":
            m = re.search(r"handler\s*=\s*(.+)\s*[,)]", source)
            if m is None:
                raise ValueError(f"Failed to find handler for {f.name}")
            source = m.group(1)
        j = {
            "project_id": project_ids.get(f.project),
            "name": f.name,
            "slug": f.slug,
            "description": f.description,
            "function_data": {
                "type": "code",
                "data": {
                    "type": "bundle",
                    "runtime_context": {
                        "runtime": "python",
                        "version": _py_version(),
                    },
                    "location": {
                        "type": "function",
                        "index": i,
                    },
                    "bundle_id": bundle_id,
                    "preview": source.strip(),
                },
            },
            "function_type": f.type_,
            "function_schema": {
                "parameters": f.parameters,
                "returns": f.returns,
            },
            "if_exists": f.if_exists if f.if_exists else if_exists,
        }
        if f.parameters is None:
            raise ValueError(f"Function {f.name} has no supplied parameters")
        j["function_schema"] = {
            "parameters": _pydantic_to_json_schema(f.parameters),
        }
        if f.returns is not None:
            j["function_schema"]["returns"] = _pydantic_to_json_schema(f.returns)
        functions.append(j)


def _collect_prompt_function_defs(
    project_ids: _ProjectIdCache, functions: List[Dict[str, Any]], if_exists: IfExists
) -> None:
    for p in global_.prompts:
        prompt_data = p.prompt
        if len(p.tool_functions) > 0:
            resolvable_tool_functions = []
            for f in p.tool_functions:
                if isinstance(f, CodeFunction):
                    resolvable_tool_functions.append(
                        {
                            "type": "slug",
                            "project_id": project_ids.get(f.project),
                            "slug": f.slug,
                        }
                    )
                else:
                    resolvable_tool_functions.append(f)
            prompt_data["tool_functions"] = resolvable_tool_functions
        j = {
            "project_id": project_ids.get(p.project),
            "name": p.name,
            "slug": p.slug,
            "function_data": {
                "type": "prompt",
            },
            "prompt_data": prompt_data,
            "if_exists": p.if_exists if p.if_exists is not None else if_exists,
        }
        if p.description is not None:
            j["description"] = p.description
        if p.function_type is not None:
            j["function_type"] = p.function_type
        functions.append(j)


def run(args):
    """Runs the braintrust push subcommand."""
    login(
        api_key=args.api_key,
        org_name=args.org_name,
        app_url=args.app_url,
    )

    if sys.path[0] != os.getcwd():
        raise ValueError(
            f"The current working directory ({os.getcwd()}) is not the project root. "
            "Please run the push command from the project root."
        )
    path = os.path.abspath(args.file)
    module_name = re.sub(".py$", "", os.path.relpath(path).replace("-", "_").replace("/", "."))

    try:
        sources = _import_module(module_name, path)
    except ImportError as e:
        if str(e) == "attempted relative import with no known parent package":
            raise ImportError(
                "Attempted to import a module using relative imports (e.g. from . import foo), but Python "
                "cannot resolve these imports without a parent package. To fix this, either: "
                "(1) combine all your code into a single file, or "
                "(2) set up a proper Python package with an __init__.py file."
            ) from e
        raise
    except Exception as e:
        raise

    project_ids = _ProjectIdCache()
    functions: List[Dict[str, Any]] = []
    if len(global_.functions) > 0:
        bundle_id = _upload_bundle(module_name, sources, args.requirements)
        _collect_function_function_defs(project_ids, functions, bundle_id, args.if_exists)
    if len(global_.prompts) > 0:
        _collect_prompt_function_defs(project_ids, functions, args.if_exists)

    if len(functions) > 0:
        api_conn().post_json("insert-functions", {"functions": functions})
    else:
        print("No functions found in the module. Nothing was pushed.", file=sys.stderr)


def build_parser(subparsers, parent_parser):
    """Adds the parser for the push subcommand."""
    parser = subparsers.add_parser(
        "push",
        help="Push code to Braintrust",
        parents=[parent_parser],
    )
    parser.add_argument(
        "--api-key",
        help="Specify a Braintrust api key. If the parameter is not specified, the BRAINTRUST_API_KEY environment variable will be used.",
    )
    parser.add_argument(
        "--org-name",
        help="The name of a specific organization to connect to. This is useful if you belong to multiple.",
    )
    parser.add_argument(
        "--app-url",
        help="Specify a custom Braintrust app url. Defaults to https://www.braintrust.dev. This is only necessary if you are using an experimental version of Braintrust.",
    )
    parser.add_argument(
        "--if-exists",
        default="error",
        choices=["error", "replace", "ignore"],
        help="What to do if a function with the same slug already exists. 'error' will cause an error and abort. 'replace' will overwrite the existing function. 'ignore' will ignore the push for this function and continue.",
    )

    parser.add_argument(
        "file",
        help="File to push.",
    )
    parser.add_argument("--requirements", help="The requirements file to bundle dependencies from.")
    parser.set_defaults(func=run)



"""
Utilities for working with audit headers.
"""
import base64
import gzip
import json
from typing import List, TypedDict


class AuditResource(TypedDict):
    type: str
    id: str
    name: str


def parse_audit_resources(hdr: str) -> List[AuditResource]:
    j = json.loads(hdr)
    if j["v"] == 1:
        return json.loads(gzip.decompress(base64.b64decode(j["p"])))
    else:
        raise ValueError(f'Unsupported audit resources protocol version: {j["v"]}')



import asyncio
import fnmatch
import importlib
import logging
import os
import sys
from dataclasses import dataclass, field
from threading import Lock
from typing import Dict, List, Optional, Union

from .. import login
from ..framework import (
    BaseExperiment,
    Evaluator,
    EvaluatorInstance,
    ReporterDef,
    _evals,
    _set_lazy_load,
    bcolors,
    default_reporter,
    init_experiment,
    parse_filters,
    report_evaluator_result,
    run_evaluator,
    set_thread_pool_max_workers,
)
from ..logger import Dataset
from ..util import eprint

INCLUDE = [
    "**/eval_*.py",
]
EXCLUDE = ["**/site-packages/**"]

_logger = logging.getLogger("braintrust.eval")


_import_lock = Lock()


@dataclass
class FileHandle:
    in_file: str

    def rebuild(self):
        in_file = os.path.abspath(self.in_file)

        with _import_lock:
            with _set_lazy_load(True):
                _evals.clear()

                try:
                    # https://stackoverflow.com/questions/67631/how-can-i-import-a-module-dynamically-given-the-full-path
                    spec = importlib.util.spec_from_file_location("eval", in_file)
                    module = importlib.util.module_from_spec(spec)
                    spec.loader.exec_module(module)

                    ret = _evals.copy()
                finally:
                    _evals.clear()

        return ret

    def watch(self):
        raise NotImplementedError


@dataclass
class EvaluatorOpts:
    verbose: bool
    no_send_logs: bool
    no_progress_bars: bool
    terminate_on_failure: bool
    watch: bool
    filters: List[str]
    list: bool
    jsonl: bool


@dataclass
class LoadedEvaluator:
    handle: FileHandle
    evaluator: Evaluator
    reporter: Optional[Union[ReporterDef, str]] = None


@dataclass
class EvaluatorState:
    evaluators: List[LoadedEvaluator] = field(default_factory=list)
    reporters: Dict[str, ReporterDef] = field(default_factory=dict)


def update_evaluators(eval_state: EvaluatorState, handles, terminate_on_failure):
    for handle in handles:
        try:
            module_evals = handle.rebuild()
        except Exception as e:
            if terminate_on_failure:
                raise
            else:
                eprint(f"Failed to import {handle.in_file}: {e}")
                continue

        for evaluator in module_evals.evaluators.values():
            if not isinstance(evaluator, EvaluatorInstance):
                continue

            eval_state.evaluators.append(
                LoadedEvaluator(handle=handle, evaluator=evaluator.evaluator, reporter=evaluator.reporter)
            )

        for reporter_name, reporter in module_evals.reporters.items():
            if not isinstance(reporter, ReporterDef):
                continue

            if reporter_name in eval_state.reporters:
                _logger.warning(
                    f"Reporter {reporter_name} already exists (in {eval_state.reporters[reporter_name].module} and {handle.in_file}). Will skip {reporter_name} in {handle.in_file}."
                )
                continue

            eval_state.reporters[reporter_name] = reporter


async def run_evaluator_task(evaluator, position, opts: EvaluatorOpts):
    experiment = None
    if not opts.no_send_logs:
        base_experiment_name = None
        if isinstance(evaluator.data, BaseExperiment):
            base_experiment_name = evaluator.data.name

        dataset = None
        if isinstance(evaluator.data, Dataset):
            dataset = evaluator.data

        # NOTE: This code is duplicated with _EvalCommon in py/src/braintrust/framework.py.
        # Make sure to update those arguments if you change this.
        experiment = init_experiment(
            project_name=evaluator.project_name,
            project_id=evaluator.project_id,
            experiment_name=evaluator.experiment_name,
            metadata=evaluator.metadata,
            is_public=evaluator.is_public,
            update=evaluator.update,
            base_experiment=base_experiment_name,
            base_experiment_id=evaluator.base_experiment_id,
            git_metadata_settings=evaluator.git_metadata_settings,
            repo_info=evaluator.repo_info,
            dataset=dataset,
        )

    try:
        return await run_evaluator(
            experiment, evaluator, position if not opts.no_progress_bars else None, opts.filters
        )
    finally:
        if experiment:
            experiment.flush()


def resolve_reporter(reporter: Optional[Union[ReporterDef, str]], reporters: Dict[str, ReporterDef]) -> ReporterDef:
    if isinstance(reporter, str):
        if reporter not in reporters:
            raise ValueError(f"Reporter {reporter} not found")
        return reporters[reporter]
    elif reporter:
        return reporter
    elif not reporters:
        return default_reporter
    elif len(reporters) == 1:
        return next(iter(reporters.values()))
    else:
        reporter_names = ", ".join(reporters.keys())
        raise ValueError(f"Multiple reporters found ({reporter_names}). Please specify a reporter explicitly.")


def add_report(eval_reports, reporter, report):
    if reporter.name not in eval_reports:
        eval_reports[reporter.name] = {"reporter": reporter, "results": []}
    eval_reports[reporter.name]["results"].append(report)


async def run_once(handles: List[FileHandle], evaluator_opts: EvaluatorOpts) -> bool:
    objects = EvaluatorState()
    update_evaluators(objects, handles, terminate_on_failure=evaluator_opts.terminate_on_failure)

    if evaluator_opts.list:
        for evaluator in objects.evaluators:
            print(f"{evaluator.evaluator.eval_name}")
        return True

    eval_promises = [
        asyncio.create_task(run_evaluator_task(evaluator.evaluator, idx, evaluator_opts))
        for idx, evaluator in enumerate(objects.evaluators)
    ]
    eval_results = [await p for p in eval_promises]

    eval_reports = {}
    for evaluator, result in zip(objects.evaluators, eval_results):
        resolved_reporter = resolve_reporter(evaluator.reporter, objects.reporters)
        report = resolved_reporter._call_report_eval(
            evaluator=evaluator.evaluator, result=result, verbose=evaluator_opts.verbose, jsonl=evaluator_opts.jsonl
        )
        add_report(eval_reports, resolved_reporter, report)

    all_success = True
    for report in eval_reports.values():
        reporter = report["reporter"]
        results = [await r for r in report["results"]]
        if not await reporter._call_report_run(results, verbose=evaluator_opts.verbose, jsonl=evaluator_opts.jsonl):
            _logger.error(f"Reporter {reporter.name} failed")
            all_success = False

    return all_success


def check_match(path_input, include_patterns, exclude_patterns):
    p = os.path.abspath(path_input)
    if include_patterns:
        include = False
        for pattern in include_patterns:
            if fnmatch.fnmatch(p, pattern):
                include = True
                break
        if not include:
            return False

    if exclude_patterns:
        exclude = False
        for pattern in exclude_patterns:
            if fnmatch.fnmatch(p, pattern):
                exclude = True
                break
        return not exclude

    return True


def collect_files(input_path):
    if os.path.isdir(input_path):
        for root, dirs, files in os.walk(input_path):
            for file in files:
                fname = os.path.join(root, file)
                if check_match(fname, INCLUDE, EXCLUDE):
                    yield fname
    else:
        if not check_match(input_path, INCLUDE, EXCLUDE):
            _logger.warning(
                f"Reading {input_path} because it was specified directly. Rename it to eval_*.py "
                + "to include it automatically when you specify a directory."
            )
        yield input_path


def initialize_handles(files):
    input_paths = files if len(files) > 0 else ["."]

    fnames = set()
    for path in input_paths:
        for fname in collect_files(path):
            fnames.add(os.path.abspath(fname))

    return [FileHandle(in_file=fname) for fname in fnames]


def run(args):
    if args.num_workers:
        set_thread_pool_max_workers(args.num_workers)

    if args.env_file:
        from dotenv import load_dotenv

        load_dotenv(args.env_file)

    evaluator_opts = EvaluatorOpts(
        verbose=args.verbose > 0,
        no_send_logs=args.no_send_logs,
        no_progress_bars=args.no_progress_bars,
        terminate_on_failure=args.terminate_on_failure,
        watch=args.watch,
        filters=parse_filters(args.filter) if args.filter else [],
        list=args.list,
        jsonl=args.jsonl,
    )

    if args.watch:
        eprint("Watch mode is not yet implemented")
        exit(1)
    if args.watch and args.list:
        eprint("Cannot specify both --list and --watch")
        exit(1)

    handles = initialize_handles(args.files)

    if not evaluator_opts.no_send_logs:
        login(
            api_key=args.api_key,
            org_name=args.org_name,
            app_url=args.app_url,
        )

    if not asyncio.run(run_once(handles, evaluator_opts)):
        sys.exit(1)


def build_parser(subparsers, parent_parser):
    parser = subparsers.add_parser(
        "eval",
        help="Run evals locally.",
        parents=[parent_parser],
    )

    parser.add_argument(
        "--api-key",
        help="Specify a braintrust api key. If the parameter is not specified, the BRAINTRUST_API_KEY environment variable will be used.",
    )
    parser.add_argument(
        "--org-name",
        help="The name of a specific organization to connect to. This is useful if you belong to multiple.",
    )
    parser.add_argument(
        "--app-url",
        help="Specify a custom braintrust app url. Defaults to https://www.braintrust.dev. This is only necessary if you are using an experimental version of Braintrust",
    )
    parser.add_argument(
        "--watch",
        action="store_true",
        help="Watch files for changes and rerun evals when changes are detected",
    )
    parser.add_argument(
        "--filter",
        help="Only run evaluators that match these filters. Each filter is a regular expression (https://docs.python.org/3/library/re.html). For example, --filter metadata.priority='^P0$' input.name='foo.*bar' will only run evaluators that have metadata.priority equal to 'P0' and input.name matching the regular expression 'foo.*bar'.",
        nargs="*",
    )
    parser.add_argument("--list", help="List, but do not execute, evaluators.", action="store_true")
    parser.add_argument(
        "--jsonl",
        help="Format score summaries as jsonl, i.e. one JSON-formatted line per summary.",
        action="store_true",
    )
    parser.add_argument(
        "--no-send-logs",
        action="store_true",
        help="Do not send logs to Braintrust. Useful for testing evaluators without uploading results.",
    )
    parser.add_argument(
        "--no-progress-bars",
        action="store_true",
        help="Do not show progress bars when processing evaluators.",
    )
    parser.add_argument(
        "--terminate-on-failure",
        action="store_true",
        help="If provided, terminates on a failing eval, instead of the default (moving onto the next one).",
    )
    parser.add_argument(
        "--env-file",
        help="A path to a .env file containing environment variables to load (via dotenv).",
    )
    parser.add_argument(
        "--num-workers",
        type=int,
        help="Specify the number of concurrent worker threads to run evals over, if they are defined as synchronous functions. Async functions will be run in the single-threaded asyncio event loop. If not specified, defaults to the number of cores on the machine.",
    )
    parser.add_argument(
        "files",
        nargs="*",
        help="A list of files or directories to run. If no files are specified, the current directory is used.",
    )

    parser.set_defaults(func=run)



"""
Do not import this file directly. See `types.py` for the classes that have a stable API.

Auto-generated file -- do not modify.
"""

from __future__ import annotations

from typing import Any, Literal, Mapping, Optional, Sequence, TypedDict, Union

from typing_extensions import NotRequired

SpanType = Literal["llm", "score", "function", "eval", "task", "tool"]


class ObjectReference(TypedDict):
    object_type: Union[
        Literal["experiment", "dataset", "prompt", "function", "prompt_session"],
        Literal["project_logs"],
    ]
    """
    Type of the object the event is originating from.
    """
    object_id: str
    """
    ID of the object the event is originating from.
    """
    id: str
    """
    ID of the original event.
    """
    _xact_id: str
    """
    Transaction ID of the original event.
    """
    created: NotRequired[Optional[str]]
    """
    Created timestamp of the original event. Used to help sort in the UI
    """


class Metadata(TypedDict):
    model: NotRequired[Optional[str]]
    """
    The model used for this example
    """


class DatasetEvent(TypedDict):
    id: str
    """
    A unique identifier for the dataset event. If you don't provide one, BrainTrust will generate one for you
    """
    _xact_id: str
    """
    The transaction id of an event is unique to the network operation that processed the event insertion. Transaction ids are monotonically increasing over time and can be used to retrieve a versioned snapshot of the dataset (see the `version` parameter)
    """
    created: str
    """
    The timestamp the dataset event was created
    """
    project_id: str
    """
    Unique identifier for the project that the dataset belongs under
    """
    dataset_id: str
    """
    Unique identifier for the dataset
    """
    input: NotRequired[Optional[Any]]
    """
    The argument that uniquely define an input case (an arbitrary, JSON serializable object)
    """
    expected: NotRequired[Optional[Any]]
    """
    The output of your application, including post-processing (an arbitrary, JSON serializable object)
    """
    metadata: NotRequired[Optional[Metadata]]
    """
    A dictionary with additional data about the test example, model outputs, or just about anything else that's relevant, that you can use to help find and analyze examples later. For example, you could log the `prompt`, example's `id`, or anything else that would be useful to slice/dice later. The values in `metadata` can be any JSON-serializable type, but its keys must be strings
    """
    tags: NotRequired[Optional[Sequence[str]]]
    """
    A list of tags to log
    """
    span_id: str
    """
    A unique identifier used to link different dataset events together as part of a full trace. See the [tracing guide](https://www.braintrust.dev/docs/guides/tracing) for full details on tracing
    """
    root_span_id: str
    """
    A unique identifier for the trace this dataset event belongs to
    """
    is_root: NotRequired[Optional[bool]]
    """
    Whether this span is a root span
    """
    origin: NotRequired[Optional[ObjectReference]]
    output: NotRequired[Optional[Any]]
    """
    Deprecated.
    """


class SpanAttributes(TypedDict):
    name: NotRequired[Optional[str]]
    """
    Name of the span, for display purposes only
    """
    type: NotRequired[Optional[SpanType]]
    """
    Type of the span, for display purposes only
    """


class Metrics(TypedDict):
    start: NotRequired[Optional[float]]
    """
    A unix timestamp recording when the section of code which produced the experiment event started
    """
    end: NotRequired[Optional[float]]
    """
    A unix timestamp recording when the section of code which produced the experiment event finished
    """
    prompt_tokens: NotRequired[Optional[int]]
    """
    The number of tokens in the prompt used to generate the experiment event (only set if this is an LLM span)
    """
    completion_tokens: NotRequired[Optional[int]]
    """
    The number of tokens in the completion generated by the model (only set if this is an LLM span)
    """
    tokens: NotRequired[Optional[int]]
    """
    The total number of tokens in the input and output of the experiment event.
    """
    caller_functionname: NotRequired[Optional[Any]]
    """
    This metric is deprecated
    """
    caller_filename: NotRequired[Optional[Any]]
    """
    This metric is deprecated
    """
    caller_lineno: NotRequired[Optional[Any]]
    """
    This metric is deprecated
    """


class Context(TypedDict):
    caller_functionname: NotRequired[Optional[str]]
    """
    The function in code which created the experiment event
    """
    caller_filename: NotRequired[Optional[str]]
    """
    Name of the file in code where the experiment event was created
    """
    caller_lineno: NotRequired[Optional[int]]
    """
    Line of code where the experiment event was created
    """


class ExperimentEvent(TypedDict):
    id: str
    """
    A unique identifier for the experiment event. If you don't provide one, BrainTrust will generate one for you
    """
    _xact_id: str
    """
    The transaction id of an event is unique to the network operation that processed the event insertion. Transaction ids are monotonically increasing over time and can be used to retrieve a versioned snapshot of the experiment (see the `version` parameter)
    """
    created: str
    """
    The timestamp the experiment event was created
    """
    project_id: str
    """
    Unique identifier for the project that the experiment belongs under
    """
    experiment_id: str
    """
    Unique identifier for the experiment
    """
    input: NotRequired[Optional[Any]]
    """
    The arguments that uniquely define a test case (an arbitrary, JSON serializable object). Later on, Braintrust will use the `input` to know whether two test cases are the same between experiments, so they should not contain experiment-specific state. A simple rule of thumb is that if you run the same experiment twice, the `input` should be identical
    """
    output: NotRequired[Optional[Any]]
    """
    The output of your application, including post-processing (an arbitrary, JSON serializable object), that allows you to determine whether the result is correct or not. For example, in an app that generates SQL queries, the `output` should be the _result_ of the SQL query generated by the model, not the query itself, because there may be multiple valid queries that answer a single question
    """
    expected: NotRequired[Optional[Any]]
    """
    The ground truth value (an arbitrary, JSON serializable object) that you'd compare to `output` to determine if your `output` value is correct or not. Braintrust currently does not compare `output` to `expected` for you, since there are so many different ways to do that correctly. Instead, these values are just used to help you navigate your experiments while digging into analyses. However, we may later use these values to re-score outputs or fine-tune your models
    """
    error: NotRequired[Optional[Any]]
    """
    The error that occurred, if any.
    """
    scores: NotRequired[Optional[Mapping[str, Optional[float]]]]
    """
    A dictionary of numeric values (between 0 and 1) to log. The scores should give you a variety of signals that help you determine how accurate the outputs are compared to what you expect and diagnose failures. For example, a summarization app might have one score that tells you how accurate the summary is, and another that measures the word similarity between the generated and grouth truth summary. The word similarity score could help you determine whether the summarization was covering similar concepts or not. You can use these scores to help you sort, filter, and compare experiments
    """
    metadata: NotRequired[Optional[Metadata]]
    """
    A dictionary with additional data about the test example, model outputs, or just about anything else that's relevant, that you can use to help find and analyze examples later. For example, you could log the `prompt`, example's `id`, or anything else that would be useful to slice/dice later. The values in `metadata` can be any JSON-serializable type, but its keys must be strings
    """
    tags: NotRequired[Optional[Sequence[str]]]
    """
    A list of tags to log
    """
    metrics: NotRequired[Optional[Metrics]]
    """
    Metrics are numerical measurements tracking the execution of the code that produced the experiment event. Use "start" and "end" to track the time span over which the experiment event was produced
    """
    context: NotRequired[Optional[Context]]
    """
    Context is additional information about the code that produced the experiment event. It is essentially the textual counterpart to `metrics`. Use the `caller_*` attributes to track the location in code which produced the experiment event
    """
    span_id: str
    """
    A unique identifier used to link different experiment events together as part of a full trace. See the [tracing guide](https://www.braintrust.dev/docs/guides/tracing) for full details on tracing
    """
    span_parents: NotRequired[Optional[Sequence[str]]]
    """
    An array of the parent `span_ids` of this experiment event. This should be empty for the root span of a trace, and should most often contain just one parent element for subspans
    """
    root_span_id: str
    """
    A unique identifier for the trace this experiment event belongs to
    """
    span_attributes: NotRequired[Optional[SpanAttributes]]
    is_root: NotRequired[Optional[bool]]
    """
    Whether this span is a root span
    """
    origin: NotRequired[Optional[ObjectReference]]


class ResponseFormat(TypedDict):
    type: Literal["json_object"]


class JsonSchema(TypedDict):
    name: str
    description: NotRequired[Optional[str]]
    schema: NotRequired[Optional[Union[Mapping[str, Any], str]]]
    strict: NotRequired[Optional[bool]]


class ResponseFormat1(TypedDict):
    type: Literal["json_schema"]
    json_schema: JsonSchema


class ResponseFormat2(TypedDict):
    type: Literal["text"]


class Function(TypedDict):
    name: str


class ToolChoice(TypedDict):
    type: Literal["function"]
    function: Function


class FunctionCall(TypedDict):
    name: str


class ModelParams1(TypedDict):
    use_cache: NotRequired[Optional[bool]]
    temperature: NotRequired[Optional[float]]
    top_p: NotRequired[Optional[float]]
    max_tokens: NotRequired[Optional[float]]
    max_completion_tokens: NotRequired[Optional[float]]
    """
    The successor to max_tokens
    """
    frequency_penalty: NotRequired[Optional[float]]
    presence_penalty: NotRequired[Optional[float]]
    response_format: NotRequired[Optional[Union[ResponseFormat, ResponseFormat1, ResponseFormat2]]]
    tool_choice: NotRequired[Optional[Union[Literal["auto"], Literal["none"], Literal["required"], ToolChoice]]]
    function_call: NotRequired[Optional[Union[Literal["auto"], Literal["none"], FunctionCall]]]
    n: NotRequired[Optional[float]]
    stop: NotRequired[Optional[Sequence[str]]]
    reasoning_effort: NotRequired[Optional[Literal["low", "medium", "high"]]]


class ModelParams2(TypedDict):
    use_cache: NotRequired[Optional[bool]]
    max_tokens: float
    temperature: float
    top_p: NotRequired[Optional[float]]
    top_k: NotRequired[Optional[float]]
    stop_sequences: NotRequired[Optional[Sequence[str]]]
    max_tokens_to_sample: NotRequired[Optional[float]]
    """
    This is a legacy parameter that should not be used.
    """


class ModelParams3(TypedDict):
    use_cache: NotRequired[Optional[bool]]
    temperature: NotRequired[Optional[float]]
    maxOutputTokens: NotRequired[Optional[float]]
    topP: NotRequired[Optional[float]]
    topK: NotRequired[Optional[float]]


class ModelParams4(TypedDict):
    use_cache: NotRequired[Optional[bool]]
    temperature: NotRequired[Optional[float]]
    topK: NotRequired[Optional[float]]


class ModelParams5(TypedDict):
    use_cache: NotRequired[Optional[bool]]


ModelParams = Union[ModelParams1, ModelParams2, ModelParams3, ModelParams4, ModelParams5]


class PromptOptions(TypedDict):
    model: NotRequired[Optional[str]]
    params: NotRequired[Optional[ModelParams]]
    position: NotRequired[Optional[str]]


class BraintrustAttachmentReference(TypedDict):
    type: Literal["braintrust_attachment"]
    """
    An identifier to help disambiguate parsing.
    """
    filename: str
    """
    Human-readable filename for user interfaces. Not related to attachment storage.
    """
    content_type: str
    """
    MIME type of this file.
    """
    key: str
    """
    Key in the object store bucket for this attachment.
    """


class ExternalAttachmentReference(TypedDict):
    type: Literal["external_attachment"]
    """
    An identifier to help disambiguate parsing.
    """
    filename: str
    """
    Human-readable filename for user interfaces. Not related to attachment storage.
    """
    content_type: str
    """
    MIME type of this file.
    """
    url: str
    """
    Fully qualified URL to the object in the external object store.
    """


AttachmentReference = Union[BraintrustAttachmentReference, ExternalAttachmentReference]


UploadStatus = Literal["uploading", "done", "error"]


class AttachmentStatus(TypedDict):
    upload_status: UploadStatus
    error_message: NotRequired[Optional[str]]
    """
    Describes the error encountered while uploading.
    """


IfExists = Literal["error", "ignore", "replace"]


class SavedFunctionId1(TypedDict):
    type: Literal["function"]
    id: str


class SavedFunctionId2(TypedDict):
    type: Literal["global"]
    name: str


SavedFunctionId = Union[SavedFunctionId1, SavedFunctionId2]


class Function1(TypedDict):
    name: str
    description: NotRequired[Optional[str]]
    parameters: NotRequired[Optional[Mapping[str, Any]]]
    strict: NotRequired[Optional[bool]]


class ToolFunctionDefinition(TypedDict):
    type: Literal["function"]
    function: Function1


class ChatCompletionContentPartText(TypedDict):
    text: NotRequired[Optional[str]]
    type: Literal["text"]


class ImageUrl(TypedDict):
    url: str
    detail: NotRequired[Optional[Union[Literal["auto"], Literal["low"], Literal["high"]]]]


class ChatCompletionContentPartImage(TypedDict):
    image_url: ImageUrl
    type: Literal["image_url"]


ChatCompletionContentPart = Union[ChatCompletionContentPartText, ChatCompletionContentPartImage]


ChatCompletionContent = Union[str, Sequence[ChatCompletionContentPart]]


class Function2(TypedDict):
    arguments: str
    name: str


class ChatCompletionMessageToolCall(TypedDict):
    id: str
    function: Function2
    type: Literal["function"]


class ChatCompletionMessageParam1(TypedDict):
    content: NotRequired[Optional[str]]
    role: Literal["system"]
    name: NotRequired[Optional[str]]


class ChatCompletionMessageParam2(TypedDict):
    content: NotRequired[Optional[ChatCompletionContent]]
    role: Literal["user"]
    name: NotRequired[Optional[str]]


class FunctionCall1(TypedDict):
    arguments: str
    name: str


class ChatCompletionMessageParam3(TypedDict):
    role: Literal["assistant"]
    content: NotRequired[Optional[str]]
    function_call: NotRequired[Optional[FunctionCall1]]
    name: NotRequired[Optional[str]]
    tool_calls: NotRequired[Optional[Sequence[ChatCompletionMessageToolCall]]]


class ChatCompletionMessageParam4(TypedDict):
    content: NotRequired[Optional[str]]
    role: Literal["tool"]
    tool_call_id: NotRequired[Optional[str]]


class ChatCompletionMessageParam5(TypedDict):
    content: NotRequired[Optional[str]]
    name: str
    role: Literal["function"]


class ChatCompletionMessageParam6(TypedDict):
    role: Literal["model"]
    content: NotRequired[Optional[str]]


ChatCompletionMessageParam = Union[
    ChatCompletionMessageParam1,
    ChatCompletionMessageParam2,
    ChatCompletionMessageParam3,
    ChatCompletionMessageParam4,
    ChatCompletionMessageParam5,
    ChatCompletionMessageParam6,
]


class Prompt(TypedDict):
    type: Literal["completion"]
    content: str


class Prompt1(TypedDict):
    type: Literal["chat"]
    messages: Sequence[ChatCompletionMessageParam]
    tools: NotRequired[Optional[str]]


class Parser(TypedDict):
    type: Literal["llm_classifier"]
    use_cot: bool
    choice_scores: Mapping[str, float]


class Origin(TypedDict):
    prompt_id: NotRequired[Optional[str]]
    project_id: NotRequired[Optional[str]]
    prompt_version: NotRequired[Optional[str]]


class PromptData(TypedDict):
    prompt: NotRequired[Optional[Union[Prompt, Prompt1]]]
    options: NotRequired[Optional[PromptOptions]]
    parser: NotRequired[Optional[Parser]]
    tool_functions: NotRequired[Optional[Sequence[SavedFunctionId]]]
    origin: NotRequired[Optional[Origin]]


__all__ = []



import unittest

from braintrust.prompt_cache import lru_cache


class TestLRUCache(unittest.TestCase):
    def test_store_and_retrieve_values(self):
        """Test storing and retrieving values."""
        cache = lru_cache.LRUCache[str, int]()
        cache.set("a", 1)
        self.assertEqual(cache.get("a"), 1)

    def test_raise_keyerror_for_missing_keys(self):
        """Test raising KeyError for missing keys."""
        cache = lru_cache.LRUCache[str, int]()
        with self.assertRaises(KeyError):
            cache.get("missing")

    def test_respect_max_size_when_specified(self):
        """Test respecting max size when specified."""
        cache = lru_cache.LRUCache[str, int](max_size=2)
        cache.set("a", 1)
        cache.set("b", 2)
        cache.set("c", 3)
        with self.assertRaises(KeyError):
            cache.get("a")
        self.assertEqual(cache.get("b"), 2)
        self.assertEqual(cache.get("c"), 3)

    def test_grow_unbounded_when_no_max_size_specified(self):
        """Test growing unbounded when no max size specified."""
        cache = lru_cache.LRUCache[int, int]()
        # Add many items.
        for i in range(1000):
            cache.set(i, i)
        # Should keep all items.
        for i in range(1000):
            self.assertEqual(cache.get(i), i)

    def test_refresh_items_on_get(self):
        """Test refreshing items on get."""
        cache = lru_cache.LRUCache[str, int](max_size=2)
        cache.set("a", 1)
        cache.set("b", 2)
        cache.get("a")  # refresh "a"
        cache.set("c", 3)
        self.assertEqual(cache.get("a"), 1)
        with self.assertRaises(KeyError):
            cache.get("b")
        self.assertEqual(cache.get("c"), 3)

    def test_update_existing_keys(self):
        """Test updating existing keys."""
        cache = lru_cache.LRUCache[str, int]()
        cache.set("a", 1)
        cache.set("a", 2)
        self.assertEqual(cache.get("a"), 2)

    def test_clear_all_items(self):
        """Test clearing all items."""
        cache = lru_cache.LRUCache[str, int]()
        cache.set("a", 1)
        cache.set("b", 2)
        cache.clear()
        with self.assertRaises(KeyError):
            cache.get("a")
        with self.assertRaises(KeyError):
            cache.get("b")


if __name__ == "__main__":
    unittest.main()



"""
A module providing a persistent disk-based cache implementation.

This module contains a generic disk cache that can store serializable objects of any type.
The cache persists entries as compressed files on disk and implements an LRU (Least Recently Used)
eviction policy based on file modification times. It provides thread-safe access to cached items
and handles file system errors gracefully.
"""

import gzip
import json
import os
from typing import Any, Callable, Generic, List, Optional, TypeVar

T = TypeVar("T")


class DiskCache(Generic[T]):
    """
    A persistent filesystem-based cache implementation.

    This cache stores entries as compressed files on disk and implements an LRU eviction
    policy based on file modification times (mtime). While access times (atime) would be more
    semantically accurate for LRU, we use mtime because:

    1. Many modern filesystems mount with noatime for performance reasons.
    2. Even when atime updates are enabled, they may be subject to update delays.
    3. mtime updates are more reliably supported across different filesystems.
    """

    def __init__(
        self,
        cache_dir: str,
        max_size: Optional[int] = None,
        serializer: Optional[Callable[[T], Any]] = None,
        deserializer: Optional[Callable[[Any], T]] = None,
    ):
        """
        Creates a new DiskCache instance.

        Args:
            cache_dir: Directory where cache files will be stored.
            max_size: Maximum number of entries to store in the cache.
                     If not specified, the cache will grow unbounded.
            serializer: Optional function to convert values to JSON-serializable format.
            deserializer: Optional function to convert JSON-deserialized data back to original type.
                         Should be the inverse of serializer.

        Example:
            # Create a cache for PromptSchema objects using its serialization methods.
            cache = DiskCache[PromptSchema](
                cache_dir="cache",
                serializer=lambda x: x.as_dict(),
                deserializer=PromptSchema.from_dict_deep
            )
        """
        self._dir = cache_dir
        self._max_size = max_size
        self._serializer = serializer
        self._deserializer = deserializer

    def _get_entry_path(self, key: str) -> str:
        """Gets the file path for a cache entry."""
        return os.path.join(self._dir, key)

    def get(self, key: str) -> T:
        """
        Retrieves a value from the cache.
        Updates the entry's access time when read.

        Args:
            key: The key to look up in the cache.

        Returns:
            The cached value.

        Raises:
            KeyError: If the key is not found in the cache.
            RuntimeError: If there is an error reading from the disk cache.
        """
        try:
            file_path = self._get_entry_path(key)
            with gzip.open(file_path, "rb") as f:
                data = json.loads(f.read().decode("utf-8"))
                if self._deserializer is not None:
                    data = self._deserializer(data)

            # Update both access and modification times.
            os.utime(file_path, None)
            return data
        except FileNotFoundError:
            raise KeyError(f"Cache key not found: {key}")
        except Exception as e:
            raise RuntimeError(f"Failed to read from disk cache: {e}") from e

    def set(self, key: str, value: T) -> None:
        """
        Stores a value in the cache.
        If the cache is at its maximum size, the least recently used entries will be evicted.

        Args:
            key: The key to store the value under.
            value: The value to store in the cache.

        Raises:
            RuntimeError: If there is an error writing to the disk cache.
        """
        try:
            os.makedirs(self._dir, exist_ok=True)
            file_path = self._get_entry_path(key)

            with gzip.open(file_path, "wb") as f:
                if self._serializer is not None:
                    value = self._serializer(value)
                f.write(json.dumps(value).encode("utf-8"))

            if self._max_size:
                entries = os.listdir(self._dir)
                if len(entries) > self._max_size:
                    self._evict_oldest(entries)
        except Exception as e:
            raise RuntimeError(f"Failed to write to disk cache: {e}") from e

    def _evict_oldest(self, entries: List[str]) -> None:
        """
        Evicts the oldest entries from the cache until it is under the maximum size.

        This method requires that self.max_size is not None, as it is only called when
        evicting entries to maintain the maximum cache size.

        Args:
            entries: List of cache entry filenames.

        Raises:
            OSError: If there is an error getting file mtimes or removing entries.
        """
        assert self._max_size is not None

        stats = []
        for entry in entries:
            path = self._get_entry_path(entry)
            mtime = os.path.getmtime(path)
            stats.append({"name": entry, "mtime": mtime})

        stats.sort(key=lambda x: x["mtime"])
        to_remove = stats[0 : len(stats) - self._max_size]

        for entry in to_remove:
            os.unlink(self._get_entry_path(entry["name"]))



import logging
from typing import Any

from braintrust import span_types
from braintrust.logger import start_span

log = logging.getLogger(__name__)


# Anthropic model parameters that we want to track as span metadata.
METADATA_PARAMS = (
    "model",
    "max_tokens",
    "temperature",
    "top_k",
    "top_p",
    "stop_sequences",
    "tool_choice",
    "tools",
)


class Wrapper:
    def __init__(self, wrapped: Any):
        self.__wrapped = wrapped

    def __getattr__(self, name: str) -> Any:
        return getattr(self.__wrapped, name)


class TracedAnthropic(Wrapper):
    def __init__(self, client):
        super().__init__(client)
        self.__client = client

    @property
    def messages(self):
        return TracedMessages(self.__client.messages)


class TracedMessages(Wrapper):
    def __init__(self, messages):
        super().__init__(messages)
        self.__messages = messages

    def stream(self, *args, **kwargs):
        return self.__trace_stream(self.__messages.stream, *args, **kwargs)

    def create(self, *args, **kwargs):
        # If stream is True, we need to trace the stream function
        if kwargs.get("stream"):
            return self.__trace_stream(self.__messages.create, *args, **kwargs)

        # Otherwise, trace synchronouly.
        _input = self.__get_input_from_kwargs(kwargs)
        metadata = self.__get_metadata_from_kwargs(kwargs)

        span = start_span(name="anthropic.messages.create", type="llm", metadata=metadata, input=_input)
        try:
            msg = self.__messages.create(*args, **kwargs)
            metrics = _extract_metrics(getattr(msg, "usage", {}))
            span.log(input=_input, output=msg.content, metrics=metrics)
            return msg
        except Exception as e:
            try:
                span.log(error=e)
            except Exception:
                pass
            raise e
        finally:
            span.end()

    def __trace_stream(self, stream_func, *args, **kwargs):
        _input = self.__get_input_from_kwargs(kwargs)
        metadata = self.__get_metadata_from_kwargs(kwargs)
        span = start_span(
            name="anthropic.messages.stream",
            metadata=metadata,
            input=_input,
            type="llm",
        )

        s = stream_func(*args, **kwargs)
        return TracedMessageStreamManager(s, span)

    @staticmethod
    def __get_input_from_kwargs(kwargs):
        msgs = list(kwargs.get("messages", []))
        # save a copy of the messages because it might be a generator
        # and we may mutate it.
        kwargs["messages"] = msgs.copy()

        system = kwargs.get("system", None)
        if system:
            msgs.append({"role": "system", "content": system})
        return msgs

    @staticmethod
    def __get_metadata_from_kwargs(kwargs):
        metadata = {"provider": "anthropic"}
        for k in METADATA_PARAMS:
            v = kwargs.get(k, None)
            if v is not None:
                metadata[k] = v
        return metadata


class TracedMessageStreamManager(Wrapper):
    def __init__(self, msg_stream_mgr, span):
        super().__init__(msg_stream_mgr)
        self.__msg_stream_mgr = msg_stream_mgr
        self.__span = span

    async def __aenter__(self):
        ms = await self.__msg_stream_mgr.__aenter__()
        return TracedMessageStream(ms, self.__span)

    def __aexit__(self, exc_type, exc_value, traceback):
        return self.__msg_stream_mgr.__aexit__(exc_type, exc_value, traceback)

    def __enter__(self):
        ms = self.__msg_stream_mgr.__enter__()
        return TracedMessageStream(ms, self.__span)

    def __exit__(self, exc_type, exc_value, traceback):
        # do we need to implement __exit__? the span is ended when the iterator is exhausted
        self.__msg_stream_mgr.__exit__(exc_type, exc_value, traceback)


class TracedMessageStream(Wrapper):
    """TracedMessageStream wraps both sync and async message streams. Obviously only one
    makes sense at a time
    """

    def __init__(self, msg_stream, span):
        super().__init__(msg_stream)
        self.__msg_stream = msg_stream
        self.__span = span
        self.__metrics = {}

    def __await__(self):
        return self.__msg_stream.__await__()

    def __aiter__(self):
        return self

    def __iter__(self):
        return self

    async def __anext__(self):
        try:
            m = await self.__msg_stream.__anext__()
            self.__process_message(m)
            return m
        except StopAsyncIteration:
            self.__span.end()
            raise

    def __next__(self):
        try:
            m = next(self.__msg_stream)
            self.__process_message(m)
            return m
        except StopIteration:
            self.__span.end()
            raise

    def __process_message(self, m):
        if m.type == "text":
            # snapshot accumulates the whole message as it streams in. We can send the whole thing
            # and updates will be dedup'ed downstream
            self.__span.log(output=str(m.snapshot))
        elif m.type in ("message_start", "message_delta", "message_stop"):
            # Parse the metrics from usage.
            # Note: For some messages it's on m.usage and others m.message.usage.
            usage = None
            if hasattr(m, "message"):
                usage = getattr(m.message, "usage", None)
            elif hasattr(m, "usage"):
                usage = m.usage

            # Not every message has every stat, but the total depends on sum of prompt & completion tokens,
            # so we need to track the current max for each value.
            if usage:
                cur_metrics = _extract_metrics(usage)
                for k, v in cur_metrics.items():
                    if v > self.__metrics.get(k, -1):
                        self.__metrics[k] = v
                self.__metrics["tokens"] = self.__metrics.get("prompt_tokens", 0) + self.__metrics.get(
                    "completion_tokens", 0
                )

            self.__span.log(metrics=self.__metrics)


def _extract_metrics(usage):
    metrics = {}
    if not usage:
        return {}

    def _save_if_exists_to(source, target=None):
        n = getattr(usage, source, None)
        if n is not None:
            metrics[target or source] = n

    _save_if_exists_to("input_tokens", "prompt_tokens")
    _save_if_exists_to("output_tokens", "completion_tokens")
    _save_if_exists_to("cache_read_input_tokens")
    _save_if_exists_to("cache_creation_input_tokens")
    metrics["tokens"] = metrics.get("prompt_tokens", 0) + metrics.get("completion_tokens", 0)

    return metrics


def wrap_anthropic(client):
    return TracedAnthropic(client)


def wrap_anthropic_client(client):
    # for backwards compatibility
    return TracedAnthropic(client)



import dataclasses
import json
from typing import Any, Callable, Dict, List, Optional, Union, overload

import slugify

from .types import (
    ChatCompletionMessageParam,
    IfExists,
    ModelParams,
    PromptData,
    PromptOptions,
    SavedFunctionId,
    ToolFunctionDefinition,
)


class _GlobalState:
    def __init__(self):
        self.functions: List[CodeFunction] = []
        self.prompts: List[CodePrompt] = []


global_ = _GlobalState()


@dataclasses.dataclass
class CodeFunction:
    """A generic callable, with metadata."""

    project: "Project"
    handler: Callable[..., Any]
    name: str
    slug: str
    type_: str
    description: Optional[str]
    parameters: Any
    returns: Any
    if_exists: Optional[IfExists]


@dataclasses.dataclass
class CodePrompt:
    """A prompt defined in code, with metadata."""

    project: "Project"
    name: str
    slug: str
    prompt: PromptData
    tool_functions: List[Union[CodeFunction, SavedFunctionId]]
    description: Optional[str]
    function_type: Optional[str]
    id: Optional[str]
    if_exists: Optional[IfExists]


class ToolBuilder:
    """Builder to create a tool in Braintrust."""

    def __init__(self, project: "Project"):
        self.project = project
        self._task_counter = 0

    def create(
        self,
        *,
        handler: Callable[..., Any],
        name: Optional[str] = None,
        slug: Optional[str] = None,
        description: Optional[str] = None,
        parameters: Any = None,
        returns: Any = None,
        if_exists: Optional[IfExists] = None,
    ) -> CodeFunction:
        """Creates a tool.

        Args:
            handler: The function that is called when the tool is used.
            name: The name of the tool.
            slug: A unique identifier for the tool.
            description: The description of the tool.
            parameters: The tool's input schema, as a Pydantic model.
            returns: The tool's output schema, as a Pydantic model.
            if_exists: What to do if the tool already exists.

        Returns:
            A handle to the created tool, that can be used in a prompt.
        """
        self._task_counter += 1
        if not name:
            if handler.__name__ and handler.__name__ != "<lambda>":
                name = handler.__name__
            else:
                name = f"Tool {self._task_counter}"
        assert name is not None
        if not slug:
            slug = slugify.slugify(name)
        f = CodeFunction(
            project=self.project,
            handler=handler,
            name=name,
            slug=slug,
            type_="tool",
            description=description,
            parameters=parameters,
            returns=returns,
            if_exists=if_exists,
        )
        global_.functions.append(f)
        return f


class PromptBuilder:
    """Builder to create a prompt in Braintrust."""

    def __init__(self, project: "Project"):
        self.project = project
        self._task_counter = 0

    @overload  # prompt only, no messages
    def create(
        self,
        *,
        name: Optional[str] = None,
        slug: Optional[str] = None,
        description: Optional[str] = None,
        id: Optional[str] = None,
        prompt: str,
        model: str,
        params: Optional[ModelParams] = None,
        tools: Optional[List[Union[CodeFunction, SavedFunctionId, ToolFunctionDefinition]]] = None,
        if_exists: Optional[IfExists] = None,
    ):
        ...

    @overload  # messages only, no prompt
    def create(
        self,
        *,
        name: Optional[str] = None,
        slug: Optional[str] = None,
        description: Optional[str] = None,
        id: Optional[str] = None,
        messages: List[ChatCompletionMessageParam],
        model: str,
        params: Optional[ModelParams] = None,
        tools: Optional[List[Union[CodeFunction, SavedFunctionId, ToolFunctionDefinition]]] = None,
        if_exists: Optional[IfExists] = None,
    ):
        ...

    def create(
        self,
        *,
        name: Optional[str] = None,
        slug: Optional[str] = None,
        description: Optional[str] = None,
        id: Optional[str] = None,
        prompt: Optional[str] = None,
        messages: Optional[List[ChatCompletionMessageParam]] = None,
        model: str,
        params: Optional[ModelParams] = None,
        tools: Optional[List[Union[CodeFunction, SavedFunctionId, ToolFunctionDefinition]]] = None,
        if_exists: Optional[IfExists] = None,
    ):
        """Creates a prompt.

        Args:
            name: The name of the prompt.
            slug: A unique identifier for the prompt.
            description: The description of the prompt.
            id: The ID of the prompt.
            prompt: The prompt text. Exactly one of prompt or messages must be provided.
            messages: The messages to send to the model. Exactly one of prompt or messages must be provided.
            model: The model to use for the prompt.
            params: The model parameters to use for the prompt.
            tools: The tools to use for the prompt.
            if_exists: What to do if the prompt already exists.
        """
        self._task_counter += 1
        if not name:
            name = f"Prompt {self._task_counter}"
        if not slug:
            slug = slugify.slugify(name)

        tool_functions: List[Union[CodeFunction, SavedFunctionId]] = []
        raw_tools: List[ToolFunctionDefinition] = []
        for tool in tools or []:
            if isinstance(tool, CodeFunction):
                tool_functions.append(tool)
            elif "type" in tool and "function" not in tool:
                # SavedFunctionId
                tool_functions.append(tool)
            else:
                # ToolFunctionDefinition
                raw_tools.append(tool)

        prompt_data: PromptData = {}
        if messages is not None:
            prompt_data["prompt"] = {
                "type": "chat",
                "messages": messages,
            }
            if len(raw_tools) > 0:
                prompt_data["prompt"]["tools"] = json.dumps(raw_tools)
        else:
            assert prompt is not None
            prompt_data["prompt"] = {
                "type": "completion",
                "content": prompt,
            }
        options: PromptOptions = {"model": model}
        if params is not None:
            options["params"] = params
        prompt_data["options"] = options

        p = CodePrompt(
            project=self.project,
            name=name,
            slug=slug,
            prompt=prompt_data,
            tool_functions=tool_functions,
            description=description,
            function_type=None,
            id=id,
            if_exists=if_exists,
        )
        global_.prompts.append(p)
        return p


class ScorerBuilder:
    """Builder to create a scorer in Braintrust."""

    def __init__(self, project: "Project"):
        self.project = project
        self._task_counter = 0

    # Code scorer.
    @overload
    def create(
        self,
        *,
        name: Optional[str] = None,
        slug: Optional[str] = None,
        description: Optional[str] = None,
        if_exists: Optional[IfExists] = None,
        handler: Callable[..., Any],
        parameters: Any,
        returns: Any = None,
    ):
        ...

    # LLM scorer with prompt.
    @overload
    def create(
        self,
        *,
        name: Optional[str] = None,
        slug: Optional[str] = None,
        description: Optional[str] = None,
        if_exists: Optional[IfExists] = None,
        prompt: str,
        model: str,
        params: Optional[ModelParams] = None,
        use_cot: bool,
        choice_scores: Dict[str, float],
    ):
        ...

    # LLM scorer with messages.
    @overload
    def create(
        self,
        *,
        name: Optional[str] = None,
        slug: Optional[str] = None,
        description: Optional[str] = None,
        if_exists: Optional[IfExists] = None,
        messages: List[ChatCompletionMessageParam],
        model: str,
        params: Optional[ModelParams] = None,
        use_cot: bool,
        choice_scores: Dict[str, float],
    ):
        ...

    def create(
        self,
        *,
        name: Optional[str] = None,
        slug: Optional[str] = None,
        description: Optional[str] = None,
        if_exists: Optional[IfExists] = None,
        # Code scorer params.
        handler: Optional[Callable[..., Any]] = None,
        parameters: Any = None,
        returns: Any = None,
        # LLM scorer params.
        prompt: Optional[str] = None,
        messages: Optional[List[ChatCompletionMessageParam]] = None,
        model: Optional[str] = None,
        params: Optional[ModelParams] = None,
        use_cot: Optional[bool] = None,
        choice_scores: Optional[Dict[str, float]] = None,
    ):
        """Creates a scorer.

        Args:
            name: The name of the scorer.
            slug: A unique identifier for the scorer.
            description: The description of the scorer.
            if_exists: What to do if the scorer already exists.

            The remaining args are mutually exclusive; that is,
            the function will only accept args from one of the following overloads.

            Code scorer:
            handler: The function that is called when the scorer is used. Required.
            parameters: The scorer's input schema, as a Pydantic model. Required.
            returns: The scorer's output schema, as a Pydantic model.

            LLM scorer:
            prompt: The prompt to use for the scorer. Either prompt or messages is required.
            messages: The messages to use for the scorer. Either prompt or messages is required.
            model: The model to use for the scorer. Required.
            params: The model parameters to use for the scorer.
            use_cot: Whether to use chain-of-thought for the scorer. Required.
            choice_scores: The scores for each choice. Required.
        """
        self._task_counter += 1
        if name is None or len(name) == 0:
            if handler and handler.__name__ and handler.__name__ != "<lambda>":
                name = handler.__name__
            else:
                name = f"Scorer {self._task_counter}"
        if slug is None or len(slug) == 0:
            slug = slugify.slugify(name)
        if handler is not None:  # code scorer
            assert parameters is not None
            f = CodeFunction(
                project=self.project,
                handler=handler,
                name=name,
                slug=slug,
                type_="scorer",
                description=description,
                parameters=parameters,
                returns=returns,
                if_exists=if_exists,
            )
            global_.functions.append(f)
        else:  # LLM scorer
            assert model is not None
            assert use_cot is not None
            assert choice_scores is not None
            prompt_data: PromptData = {}
            if messages is not None:
                assert prompt is None
                prompt_data["prompt"] = {
                    "type": "chat",
                    "messages": messages,
                }
            else:
                assert prompt is not None
                prompt_data["prompt"] = {
                    "type": "completion",
                    "content": prompt,
                }
            prompt_data["options"] = {"model": model}
            if params is not None:
                prompt_data["options"]["params"] = params
            prompt_data["parser"] = {
                "type": "llm_classifier",
                "use_cot": use_cot,
                "choice_scores": choice_scores,
            }
            p = CodePrompt(
                project=self.project,
                name=name,
                slug=slug,
                prompt=prompt_data,
                tool_functions=[],
                description=description,
                function_type="scorer",
                id=None,
                if_exists=if_exists,
            )
            global_.prompts.append(p)


class Project:
    """A handle to a Braintrust project."""

    def __init__(self, name: str):
        self.name = name
        self.tools = ToolBuilder(self)
        self.prompts = PromptBuilder(self)
        self.scorers = ScorerBuilder(self)


class ProjectBuilder:
    """Creates handles to Braintrust projects."""

    def create(self, name: str) -> Project:
        return Project(name)


projects = ProjectBuilder()



import inspect
import os.path
import sys
import threading
import urllib.parse
from dataclasses import dataclass
from typing import Any, Callable, Dict, Generic, Literal, Mapping, Optional, Set, Tuple, TypedDict, TypeVar, Union

from requests import HTTPError, Response

GLOBAL_PROJECT = "Global"
BT_IS_ASYNC_ATTRIBUTE = "_BT_IS_ASYNC"


# Taken from
# https://stackoverflow.com/questions/5574702/how-do-i-print-to-stderr-in-python.
def eprint(*args, **kwargs) -> None:
    print(*args, file=sys.stderr, **kwargs)


def coalesce(*args):
    """Returns the first non-None value in the list of `args`, or `None` if they
    are all `None`.
    """

    for a in args:
        if a is not None:
            return a
    return None


def merge_dicts_with_paths(
    merge_into: Dict[str, Any], merge_from: Mapping[str, Any], path: Tuple[str, ...], merge_paths: Set[Tuple[str]]
) -> Dict[str, Any]:
    """Merges merge_from into merge_into, destructively updating merge_into. Does not merge any further than
    merge_paths."""

    if not isinstance(merge_into, dict):
        raise ValueError("merge_into must be a dictionary")
    if not isinstance(merge_from, dict):
        raise ValueError("merge_from must be a dictionary")

    for k, merge_from_v in merge_from.items():
        full_path = path + (k,)
        merge_into_v = merge_into.get(k)
        if isinstance(merge_into_v, dict) and isinstance(merge_from_v, dict) and full_path not in merge_paths:
            merge_dicts_with_paths(merge_into_v, merge_from_v, full_path, merge_paths)
        else:
            merge_into[k] = merge_from_v

    return merge_into


def merge_dicts(merge_into: Dict[str, Any], merge_from: Mapping[str, Any]) -> Dict[str, Any]:
    """Merges merge_from into merge_into, destructively updating merge_into."""

    return merge_dicts_with_paths(merge_into, merge_from, (), set())


def encode_uri_component(name: str) -> str:
    """Encode a single component of a URI. Slashes are encoded as well, so this
    should not be used for multiple slash-separated URI components."""

    return urllib.parse.quote(name, safe="")


def _urljoin(*parts: str) -> str:
    return "/".join(
        p for p in [x.strip("/") if i < len(parts) - 1 else x.lstrip("/") for i, x in enumerate(parts)] if p.strip()
    )


class AugmentedHTTPError(Exception):
    pass


def response_raise_for_status(resp: Response) -> None:
    try:
        resp.raise_for_status()
    except HTTPError as e:
        raise AugmentedHTTPError(f"{resp.text}") from e


class CallerLocation(TypedDict):
    caller_functionname: str
    caller_filename: str
    caller_lineno: int


def get_caller_location() -> Optional[CallerLocation]:
    # Modified from
    # https://stackoverflow.com/questions/24438976/debugging-get-filename-and-line-number-from-which-a-function-is-called
    # to fetch the first stack frame not contained inside the same directory as
    # this file.
    this_dir = None
    call_stack = inspect.stack()
    for frame in call_stack:
        caller = inspect.getframeinfo(frame.frame)
        if this_dir is None:
            this_dir = os.path.dirname(caller.filename)
        if os.path.dirname(caller.filename) != this_dir:
            return CallerLocation(
                caller_functionname=caller.function,
                caller_filename=caller.filename,
                caller_lineno=caller.lineno,
            )
    return None


T = TypeVar("T")


@dataclass
class _LazyValueResolvedState(Generic[T]):
    value: T
    has_succeeded: Literal[True] = True


@dataclass
class _LazyValuePendingState:
    has_succeeded: Literal[False] = False


_LazyValueState = Union[_LazyValueResolvedState[T], _LazyValuePendingState]


class LazyValue(Generic[T]):
    """A simple wrapper around a callable object which computes the value
    on-demand and saves it for future retrievals.
    """

    def __init__(self, callable: Callable[[], T], use_mutex: bool):
        self.callable = callable
        self.mutex = threading.Lock() if use_mutex else None
        self._state: _LazyValueState[T] = _LazyValuePendingState()

    @property
    def has_succeeded(self) -> bool:
        return self._state.has_succeeded

    @property
    def value(self) -> Optional[T]:
        return self._state.value if self._state.has_succeeded == True else None

    def get(self) -> T:
        # Short-circuit check `has_succeeded`. This should be fine because
        # setting `_state` is atomic and python should have sequentially
        # consistent semantics, so we'll observe the write to
        # `self._state.value` as well.
        # https://docs.python.org/3/faq/library.html#what-kinds-of-global-value-mutation-are-thread-safe
        if self._state.has_succeeded == True:
            return self._state.value
        if self.mutex:
            self.mutex.acquire()
        try:
            if self._state.has_succeeded == False:
                res = self.callable()
                self._state = _LazyValueResolvedState(value=res)
            return self._state.value
        finally:
            if self.mutex:
                self.mutex.release()


_MARK_ASYNC_WRAPPER_UNDERLYING_CALLABLE_ATTRIBUTE = "_MarkAsyncWrapper_underlying_callable"


# A wrapper class to enable explicitly marking a callable object as async. This
# can be useful for scenarios where the user wants to provide an awaitable
# function that is not recognized as async with `inspect.iscoroutinefunction`.
#
# Note: Python 3.12 provides a `inspect.markcoroutinefunction` function which
# serves a similar purpose, but we do this ourselves in case this function is
# not available.
class MarkAsyncWrapper:
    def __init__(self, callable):
        setattr(self, _MARK_ASYNC_WRAPPER_UNDERLYING_CALLABLE_ATTRIBUTE, callable)
        setattr(self, BT_IS_ASYNC_ATTRIBUTE, True)

    def __getattribute__(self, name):
        if name in [_MARK_ASYNC_WRAPPER_UNDERLYING_CALLABLE_ATTRIBUTE, BT_IS_ASYNC_ATTRIBUTE]:
            return object.__getattribute__(self, name)
        else:
            return object.__getattribute__(
                object.__getattribute__(self, _MARK_ASYNC_WRAPPER_UNDERLYING_CALLABLE_ATTRIBUTE), name
            )

    def __call__(self, *args, **kwargs):
        return object.__getattribute__(self, _MARK_ASYNC_WRAPPER_UNDERLYING_CALLABLE_ATTRIBUTE)(*args, **kwargs)


def bt_iscoroutinefunction(f):
    return inspect.iscoroutinefunction(f) or getattr(f, BT_IS_ASYNC_ATTRIBUTE, False)


def add_azure_blob_headers(headers: Dict[str, str], url: str) -> None:
    # According to https://stackoverflow.com/questions/37824136/put-on-sas-blob-url-without-specifying-x-ms-blob-type-header,
    # there is no way to avoid including this.
    if "blob.core.windows.net" in url:
        headers["x-ms-blob-type"] = "BlockBlob"



import logging
import os
import textwrap
import time

from botocore.exceptions import ClientError

from braintrust.logger import app_conn, login

# pylint: disable=no-name-in-module
from ...aws import cloudformation
from ...util import response_raise_for_status

_logger = logging.getLogger("braintrust.install.api")

PARAMS = {
    "OrgName": "org_name",
    "ProvisionedConcurrency": "provisioned_concurrency",
    "EncryptDatabase": "encrypt_database",
    "PostgresAlternativeHost": "postgres_alternative_host",
    "APIHandlerMemorySize": "api_handler_memory_size",
    "WhitelistedOrigins": "whitelisted_origins",
    "PublicSubnet1AZ": "public_subnet_1_az",
    "PrivateSubnet1AZ": "private_subnet_1_az",
    "PrivateSubnet2AZ": "private_subnet_2_az",
    "PrivateSubnet3AZ": "private_subnet_3_az",
    "VPCCIDR": "vpc_cidr",
    "PublicSubnet1CIDR": "public_subnet_1_cidr",
    "PrivateSubnet1CIDR": "private_subnet_1_cidr",
    "PrivateSubnet2CIDR": "private_subnet_2_cidr",
    "PrivateSubnet3CIDR": "private_subnet_3_cidr",
    "ManagedPostgres": "managed_postgres",
    "ManagedClickhouse": "managed_clickhouse",
    "ClickhouseInstanceType": "clickhouse_instance_type",
    "PostgresVersion": "postgres_version",
    "OutboundRateLimitWindowMinutes": "outbound_rate_limit_window_minutes",
    "OutboundRateLimitMaxRequests": "outbound_rate_limit_max_requests",
    "UseGlobalProxy": "use_global_proxy",
    "EnableQuarantine": "enable_quarantine",
    "EnableBrainstore": "enable_brainstore",
    "BrainstoreInstanceKeyPairName": "brainstore_instance_key_pair_name",
    "BrainstoreInstanceType": "brainstore_instance_type",
    "BrainstoreInstanceCount": "brainstore_instance_count",
    "BrainstoreMaxInstanceCount": "brainstore_max_instance_count",
    "BrainstoreVersionOverride": "brainstore_version_override",
    "BrainstoreLicenseKey": "brainstore_license_key",
}

REMOVED_PARAMS = ["ThirdAZIndex"]

DEFAULTS = {
    "DwType": "Postgres",
    "EncryptDatabase": "false",
    "ProvisionedConcurrency": 0,
    "APIHandlerMemorySize": 10240,
}

CAPABILITIES = ["CAPABILITY_IAM", "CAPABILITY_AUTO_EXPAND"]

LATEST_TEMPLATE = "https://braintrust-cf.s3.amazonaws.com/braintrust-latest.yaml"


def build_parser(subparsers, parents):
    parser = subparsers.add_parser("api", help="Install the Braintrust function API", parents=parents)

    parser.add_argument("name", help="Name of the CloudFormation stack to create or update")
    parser.add_argument(
        "--create",
        help="Create the stack if it does not exist",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--vpc-connect",
        help="Connect to an existing VPC",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--template",
        help="A specific CloudFormation template URL to use",
        default=None,
    )
    parser.add_argument(
        "--update-template",
        help="Update the CloudFormation to the latest version of the template",
        action="store_true",
        default=False,
    )

    # OrgName, ProvisionedConcurrency
    parser.add_argument("--org-name", help="The name of your organization", default=None)
    parser.add_argument(
        "--provisioned-concurrency",
        help="The amount of provisioned concurrency",
        default=None,
        type=int,
    )
    parser.add_argument(
        "--api-handler-memory-size",
        help="The amount of memory to allocate to the API handler",
        default=None,
        type=int,
    )
    parser.add_argument(
        "--whitelisted-origins",
        help="Comma-separated list of origins to whitelist",
        default=None,
    )
    parser.add_argument(
        "--public-subnet-1-az",
        help="The availability zone for the public subnet",
        default=None,
    )
    parser.add_argument(
        "--private-subnet-1-az",
        help="The availability zone for private subnet 1",
        default=None,
    )
    parser.add_argument(
        "--private-subnet-2-az",
        help="The availability zone for private subnet 2",
        default=None,
    )
    parser.add_argument(
        "--private-subnet-3-az",
        help="The availability zone for private subnet 3",
        default=None,
    )

    parser.add_argument(
        "--vpc-cidr",
        help="The CIDR for the VPC",
        default=None,
    )
    parser.add_argument(
        "--public-subnet-1-cidr",
        help="The CIDR for the public subnet",
        default=None,
    )
    parser.add_argument(
        "--private-subnet-1-cidr",
        help="The CIDR for private subnet 1",
        default=None,
    )
    parser.add_argument(
        "--private-subnet-2-cidr",
        help="The CIDR for private subnet 2",
        default=None,
    )
    parser.add_argument(
        "--private-subnet-3-cidr",
        help="The CIDR for private subnet 3",
        default=None,
    )

    # PostgresUrl
    parser.add_argument(
        "--managed-postgres",
        help="Spin up an RDS instance to use as the datastore",
        default=None,
        choices=[None, "true", "false"],
    )
    parser.add_argument(
        "--encrypt-database",
        help="Whether to encrypt the database",
        default="false",
        choices=[None, "true", "false"],
    )
    parser.add_argument(
        "--postgres-version",
        help="The version of the postgres instance",
        default=None,
    )
    parser.add_argument(
        "--postgres-alternative-host",
        help="Use an external host for postgres (but the same secrets)",
        default=None,
    )

    # Clickhouse
    parser.add_argument(
        "--managed-clickhouse",
        help="Spin up a Clickhouse Instance for faster analytics",
        default=None,
        choices=[None, "true", "false"],
    )
    parser.add_argument(
        "--clickhouse-instance-type",
        help="The instance type for the Clickhouse instance",
        default=None,
    )

    # ElastiCacheClusterId
    parser.add_argument("--elasticache-cluster-host", help="The ElastiCacheCluster host to use", default=None)
    parser.add_argument(
        "--elasticache-cluster-port", help="The ElastiCacheCluster host to use", default=None, type=int
    )

    # SecurityGroupId, SubnetIds
    parser.add_argument("--security-group-id", help="The security group ID to use", default=None)
    parser.add_argument("--subnet-ids", help="The subnet IDs to use", default=None)

    # Rate limit configuration.
    parser.add_argument(
        "--outbound-rate-limit-window-minutes",
        help="The time frame in minutes over which rate per-user rate limits are accumulated",
        default=None,
        type=int,
    )
    parser.add_argument(
        "--outbound-rate-limit-max-requests",
        help="The maximum number of requests per user allowed in the time frame specified by OutboundRateLimitMaxRequests. Setting to 0 will disable rate limits",
        default=None,
        type=int,
    )

    parser.add_argument(
        "--use-global-proxy",
        help="Use the global cloudflare proxy (https://braintrustproxy.com)",
        default=None,
        choices=[None, "true", "false"],
    )

    parser.add_argument(
        "--enable-quarantine",
        help="Enable the quarantine feature (running typescript and python functions)",
        default=None,
        choices=[None, "true", "false"],
    )

    # Advancd use only
    parser.add_argument(
        "--postgres-url",
        help="[Advanced] The postgres URL to use (if you are connecting to another VPC)",
        default=None,
    )
    parser.add_argument("--clickhouse-pg-url", help="[Advanced] The clickhouse PG URL to use", default=None)
    parser.add_argument("--clickhouse-connect-url", help="[Advanced] The clickhouse connect URL to use", default=None)
    parser.add_argument(
        "--clickhouse-catchup-etl-arn", help="[Advanced] The clickhouse catchup ETL ARN to use", default=None
    )

    # To configure your org
    parser.add_argument(
        "--api-key",
        help="The API key to use to configure your org's API URL and Proxy URL",
        default=os.environ.get("BRAINTRUST_API_KEY", None),
    )

    # Brainstore configuration
    parser.add_argument(
        "--enable-brainstore",
        help="Enable Brainstore object-storage data backend",
        choices=[None, "true", "false"],
        default=None,
    )
    parser.add_argument(
        "--brainstore-license-key",
        help="The license key to use for Brainstore",
        default=None,
    )
    parser.add_argument(
        "--brainstore-instance-key-pair-name",
        help="The EC2 Key Pair to allow SSH access to the Brainstore instance",
        default=None,
    )
    parser.add_argument(
        "--brainstore-instance-type",
        help="EC2 instance type for Brainstore. Must be a Graviton instance type.",
        default=None,
    )
    parser.add_argument(
        "--brainstore-instance-count",
        help="Number of Brainstore instances to run",
        type=int,
        default=None,
    )
    parser.add_argument(
        "--brainstore-max-instance-count",
        help="Max scaling size for Brainstore instances",
        type=int,
        default=None,
    )
    parser.add_argument(
        "--brainstore-version-override",
        help="Lock Brainstore to a specific docker tag",
        default=None,
    )

    parser.set_defaults(func=main)


def main(args):
    template = args.template or LATEST_TEMPLATE

    status = None
    try:
        statuses = cloudformation.describe_stacks(StackName=args.name)["Stacks"]
        if len(statuses) == 1:
            status = statuses[0]
        _logger.debug(status)
    except ClientError as e:
        if "does not exist" not in str(e):
            raise

    vpc_connect = args.vpc_connect
    if status and not vpc_connect:
        vpc_connect = "SecurityGroupId" in set(x["ParameterKey"] for x in status["Parameters"])

    if vpc_connect:
        PARAMS["SecurityGroupId"] = "security_group_id"
        PARAMS["SubnetIds"] = "subnet_ids"
        PARAMS["ElastiCacheClusterHost"] = "elasticache_cluster_host"
        PARAMS["ElastiCacheClusterPort"] = "elasticache_cluster_port"
        PARAMS["PostgresUrl"] = "postgres_url"
        PARAMS["ClickhouseCatchupEtlArn"] = "clickhouse_catchup_etl_arn"
        PARAMS["ClickhouseConnectUrl"] = "clickhouse_connect_url"
        PARAMS["ClickhousePGUrl"] = "clickhouse_pg_url"
        PARAMS["EnableBrainstore"] = "enable_brainstore"
        PARAMS["BrainstoreInstanceKeyPairName"] = "brainstore_instance_key_pair_name"
        PARAMS["BrainstoreLicenseKey"] = "brainstore_license_key"
        PARAMS["BrainstoreInstanceType"] = "brainstore_instance_type"
        PARAMS["BrainstoreInstanceCount"] = "brainstore_instance_count"
        PARAMS["BrainstoreMaxInstanceCount"] = "brainstore_max_instance_count"
        PARAMS["BrainstoreVersionOverride"] = "brainstore_version_override"

        if args.template is None:
            template = "https://braintrust-cf.s3.amazonaws.com/braintrust-latest-vpc.yaml"

    exists = status is not None
    if exists and args.create:
        _logger.error(
            textwrap.dedent(
                f"""\
            Stack with name {args.name} already exists. Either delete it in the AWS console or
            remove the --create flag."""
            )
        )
        exit(1)
    elif not exists and not args.create:
        _logger.error(
            textwrap.dedent(
                f"""\
            Stack with name {args.name} does not exist. Either create it manually by following
            https://www.braintrust.dev/docs/guides/self-hosting/aws or use the --create flag."""
            )
        )
        exit(1)

    if not exists:
        _logger.info(f"Creating stack with name {args.name}")

        params = [
            {
                "ParameterKey": k,
                "ParameterValue": str(v),
            }
            for (k, v) in [
                (param, args.__dict__[arg_name] or DEFAULTS.get(param, None)) for (param, arg_name) in PARAMS.items()
            ]
            if v is not None
        ]
        _logger.info("Using params:")
        for param in params:
            _logger.info(f"  {param['ParameterKey']}: {param['ParameterValue']}")

        _logger.info(f"Typical stack creation takes 10-15 minutes.")

        cloudformation.create_stack(
            StackName=args.name,
            TemplateURL=template,
            Parameters=params,
            Capabilities=CAPABILITIES,
        )

        for _ in range(80):
            status = cloudformation.describe_stacks(StackName=args.name)["Stacks"][0]
            if status["StackStatus"] != "CREATE_IN_PROGRESS":
                exists = True
                break
            _logger.info("Waiting for stack to be created...")
            time.sleep(15)
        else:
            _logger.error(
                textwrap.dedent(
                    """\
                Stack creation timed out. Please check the AWS console to see its status. You can also
                re-run this command without --create to continue the setup process once it's done."""
                )
            )
            exit(1)
        _logger.info(f"Stack with name {args.name} has been created with status: {status['StackStatus']}")
        exit(0)

    _logger.info(f"Stack with name {args.name} has status: {status['StackStatus']}")

    if not ("_COMPLETE" in status["StackStatus"] or "_FAILED" in status["StackStatus"]):
        _logger.info(f"Please re-run this command once the stack has finished creating or updating")
        exit(0)

    # Update params that have changed
    param_updates = {}
    for param, arg_name in PARAMS.items():
        if args.__dict__[arg_name] is not None:
            param_updates[param] = args.__dict__[arg_name]
    if len(param_updates) > 0 or args.update_template:
        template_kwargs = {"TemplateURL": template} if args.update_template else {"UsePreviousTemplate": True}
        _logger.info(f"Updating stack with name {args.name} with template: {template_kwargs}")

        _logger.info("Using params:")
        for param, value in param_updates.items():
            _logger.info(f"  {param}: {value}")

        if args.template:
            new_template = cloudformation.get_template_summary(TemplateURL=template)
            new_params = set(x["ParameterKey"] for x in new_template["Parameters"])
        else:
            new_params = set(x["ParameterKey"] for x in status["Parameters"])

        stack = cloudformation.describe_stacks(StackName=args.name)["Stacks"][0]
        try:
            final_params = [
                {"ParameterKey": param, "ParameterValue": str(update)}
                for (param, update) in param_updates.items()
                if param in new_params and param not in REMOVED_PARAMS
            ] + [
                {"ParameterKey": param["ParameterKey"], "UsePreviousValue": True}
                for param in stack["Parameters"]
                if param["ParameterKey"] not in param_updates
                and param["ParameterKey"] not in REMOVED_PARAMS
                and param["ParameterKey"] in new_params
            ]
            cloudformation.update_stack(
                StackName=args.name,
                Parameters=final_params,
                Capabilities=CAPABILITIES,
                **template_kwargs,
            )
        except ClientError as e:
            if "No updates are to be performed." in str(e):
                _logger.warning("No updates are to be performed.")
            else:
                raise

        for _ in range(120):
            status = cloudformation.describe_stacks(StackName=args.name)["Stacks"][0]
            if status["StackStatus"] != "UPDATE_IN_PROGRESS":
                exists = True
                break
            _logger.info("Waiting for stack to be updated...")
            time.sleep(5)
        else:
            _logger.error(
                textwrap.dedent(
                    """\
                Stack update timed out. Please check the AWS console to see its status. You can also
                re-run this command to try again."""
                )
            )
            exit(1)

        universal_url = [x for x in status["Outputs"] if x["OutputKey"] == "UniversalURL"]
        if universal_url:
            universal_url = universal_url[0]["OutputValue"]
        else:
            universal_url = None

        org_name = [x for x in status["Parameters"] if x["ParameterKey"] == "OrgName"]
        if org_name:
            org_name = org_name[0]["ParameterValue"]
        else:
            org_name = None

        _logger.info(f"Stack with name {args.name} has been updated with status: {status['StackStatus']}")
        _logger.info(f"Universal URL: {universal_url}")

        org_info = []
        if args.api_key:
            login(api_key=args.api_key)
            resp = app_conn().post("api/apikey/login")
            if resp.ok:
                org_info = resp.json()["org_info"]
            else:
                _logger.error(f"Failed to login with API key: {resp.text}")

        if len(org_info) > 0:
            if org_name != "*":
                org_info = [x for x in org_info if x["name"] == org_name]
            if len(org_info) == 0:
                _logger.error(f"Org with name {org_name} does not exist")
                exit(1)
            elif len(org_info) > 1:
                names = ", ".join([x["name"] for x in org_info])
                _logger.error(
                    f"You belong to multiple orgs: {names}. Please use an API key that's scoped to a single org."
                )
                org_info = []

            if len(org_info) == 1:
                org_info = org_info[0]

        if org_info and (universal_url and org_info["api_url"] != universal_url):
            _logger.info(f"Will update org {org_info['name']}'s urls.")
            _logger.info(f"  They are currently set to:")
            _logger.info(f"  API URL: {org_info['api_url']}")
            _logger.info(f"  Proxy URL: {org_info['proxy_url']}")
            _logger.info(f"And will update them to:")

            patch_args = {"id": org_info["id"]}
            if universal_url and org_info["api_url"] != universal_url:
                patch_args["api_url"] = universal_url
                patch_args["is_universal_api"] = True
                _logger.info(f"  API URL: {universal_url}")
                _logger.warn(
                    f"\nNOTE: You can delete the proxy URL from your org settings now. It is no longer needed."
                )

            # Make the actual request
            response_raise_for_status(
                app_conn().post(
                    "api/organization/patch_id",
                    json=patch_args,
                )
            )



from typing import List
from unittest import TestCase

import braintrust
from braintrust import Attachment, BaseAttachment, ExternalAttachment, LazyValue, Prompt
from braintrust.logger import _deep_copy_event, _extract_attachments
from braintrust.prompt import PromptChatBlock, PromptData, PromptMessage, PromptSchema


class TestInit(TestCase):
    def test_init_validation(self):
        with self.assertRaises(ValueError) as cm:
            braintrust.init()

        assert str(cm.exception) == "Must specify at least one of project or project_id"

        with self.assertRaises(ValueError) as cm:
            braintrust.init(project="project", open=True, update=True)

        assert str(cm.exception) == "Cannot open and update an experiment at the same time"

        with self.assertRaises(ValueError) as cm:
            braintrust.init(project="project", open=True)

        assert str(cm.exception) == "Cannot open an experiment without specifying its name"


class TestLogger(TestCase):
    def test_extract_attachments_no_op(self):
        attachments: List[BaseAttachment] = []

        _extract_attachments({}, attachments)
        self.assertEqual(len(attachments), 0)

        event = {"foo": "foo", "bar": None, "baz": [1, 2, 3]}
        _extract_attachments(event, attachments)
        self.assertEqual(len(attachments), 0)
        # Same instance
        self.assertIs(event["baz"], event["baz"])
        # Same content
        self.assertEqual(event, {"foo": "foo", "bar": None, "baz": [1, 2, 3]})

    def test_extract_attachments_with_attachments(self):
        attachment1 = Attachment(
            data=b"data",
            filename="filename",
            content_type="text/plain",
        )
        attachment2 = Attachment(
            data=b"data2",
            filename="filename2",
            content_type="text/plain",
        )
        attachment3 = ExternalAttachment(
            url="s3://bucket/path/to/key.pdf",
            filename="filename3",
            content_type="application/pdf",
        )
        date = "2024-10-23T05:02:48.796Z"
        event = {
            "foo": "bar",
            "baz": [1, 2],
            "attachment1": attachment1,
            "attachment3": attachment3,
            "nested": {
                "attachment2": attachment2,
                "attachment3": attachment3,
                "info": "another string",
                "anArray": [
                    attachment1,
                    None,
                    "string",
                    attachment2,
                    attachment1,
                    attachment3,
                    attachment3,
                ],
            },
            "null": None,
            "undefined": None,
            "date": date,
            "f": "Math.max",
            "empty": {},
        }
        saved_nested = event["nested"]

        attachments: List[BaseAttachment] = []
        _extract_attachments(event, attachments)

        self.assertEqual(
            attachments,
            [
                attachment1,
                attachment3,
                attachment2,
                attachment3,
                attachment1,
                attachment2,
                attachment1,
                attachment3,
                attachment3,
            ],
        )
        self.assertIs(attachments[0], attachment1)
        self.assertIs(attachments[1], attachment3)
        self.assertIs(attachments[2], attachment2)
        self.assertIs(attachments[3], attachment3)
        self.assertIs(attachments[4], attachment1)
        self.assertIs(attachments[5], attachment2)
        self.assertIs(attachments[6], attachment1)
        self.assertIs(attachments[7], attachment3)
        self.assertIs(attachments[8], attachment3)

        self.assertIs(event["nested"], saved_nested)

        self.assertEqual(
            event,
            {
                "foo": "bar",
                "baz": [1, 2],
                "attachment1": attachment1.reference,
                "attachment3": attachment3.reference,
                "nested": {
                    "attachment2": attachment2.reference,
                    "attachment3": attachment3.reference,
                    "info": "another string",
                    "anArray": [
                        attachment1.reference,
                        None,
                        "string",
                        attachment2.reference,
                        attachment1.reference,
                        attachment3.reference,
                        attachment3.reference,
                    ],
                },
                "null": None,
                "undefined": None,
                "date": date,
                "f": "Math.max",
                "empty": {},
            },
        )

    def test_deep_copy_event_basic(self):
        original = {
            "input": {"foo": "bar", "null": None, "empty": {}},
            "output": [1, 2, "3", None, {}],
        }
        copy = _deep_copy_event(original)
        self.assertEqual(copy, original)
        self.assertIsNot(copy, original)
        self.assertIsNot(copy["input"], original["input"])
        self.assertIsNot(copy["output"], original["output"])

    def test_deep_copy_event_with_attachments(self):
        attachment1 = Attachment(
            data=b"data",
            filename="filename",
            content_type="text/plain",
        )
        attachment2 = Attachment(
            data=b"data2",
            filename="filename2",
            content_type="text/plain",
        )
        attachment3 = ExternalAttachment(
            url="s3://bucket/path/to/key.pdf",
            filename="filename3",
            content_type="application/pdf",
        )
        date = "2024-10-23T05:02:48.796Z"

        original = {
            "input": "Testing",
            "output": {
                "span": "<span>",
                "myIllegalObjects": ["<experiment>", "<dataset>", "<logger>"],
                "myOtherWeirdObjects": [None, date, None, None],
                "attachment": attachment1,
                "another_attachment": attachment3,
                "attachmentList": [attachment1, attachment2, "string", attachment3],
                "nestedAttachment": {
                    "attachment": attachment2,
                    "another_attachment": attachment3,
                },
                "fake": {
                    "_bt_internal_saved_attachment": "not a number",
                },
            },
        }

        copy = _deep_copy_event(original)

        self.assertEqual(
            copy,
            {
                "input": "Testing",
                "output": {
                    "span": "<span>",
                    "myIllegalObjects": ["<experiment>", "<dataset>", "<logger>"],
                    "myOtherWeirdObjects": [None, date, None, None],
                    "attachment": attachment1,
                    "another_attachment": attachment3,
                    "attachmentList": [attachment1, attachment2, "string", attachment3],
                    "nestedAttachment": {
                        "attachment": attachment2,
                        "another_attachment": attachment3,
                    },
                    "fake": {
                        "_bt_internal_saved_attachment": "not a number",
                    },
                },
            },
        )

        self.assertIsNot(copy, original)

        self.assertIs(copy["output"]["attachment"], attachment1)
        self.assertIs(copy["output"]["another_attachment"], attachment3)
        self.assertIs(copy["output"]["nestedAttachment"]["attachment"], attachment2)
        self.assertIs(copy["output"]["nestedAttachment"]["another_attachment"], attachment3)
        self.assertIs(copy["output"]["attachmentList"][0], attachment1)
        self.assertIs(copy["output"]["attachmentList"][1], attachment2)
        self.assertIs(copy["output"]["attachmentList"][3], attachment3)

    def test_prompt_build_with_structured_output_templating(self):
        self.maxDiff = None
        prompt = Prompt(
            LazyValue(
                lambda: PromptSchema(
                    id="id",
                    project_id="project_id",
                    _xact_id="_xact_id",
                    name="name",
                    slug="slug",
                    description="description",
                    prompt_data=PromptData(
                        prompt=PromptChatBlock(
                            messages=[
                                PromptMessage(
                                    role="system",
                                    content="Please compute {{input.expression}} and return the result in JSON.",
                                ),
                            ],
                        ),
                        options={
                            "model": "gpt-4o",
                            "params": {
                                "response_format": {
                                    "type": "json_schema",
                                    "json_schema": {
                                        "name": "schema",
                                        "schema": "{{input.schema}}",
                                        "strict": True,
                                    },
                                },
                            },
                        },
                    ),
                    tags=None,
                ),
                use_mutex=True,
            ),
            {},
            False,
        )

        result = prompt.build(
            **{
                "input": {
                    "expression": "2 + 3",
                    "schema": {
                        "type": "object",
                        "properties": {
                            "final_answer": {
                                "type": "string",
                            },
                        },
                        "required": ["final_answer"],
                        "additionalProperties": False,
                    },
                },
            }
        )

        self.assertEqual(
            result["response_format"],
            {
                "type": "json_schema",
                "json_schema": {
                    "name": "schema",
                    "schema": {
                        "type": "object",
                        "properties": {
                            "final_answer": {"type": "string"},
                        },
                        "required": ["final_answer"],
                        "additionalProperties": False,
                    },
                    "strict": True,
                },
            },
        )


def test_noop_permalink_issue_1837():
    # fixes issue #BRA-1837
    span = braintrust.NOOP_SPAN
    assert span.permalink() == "https://braintrust.dev/noop-span"

    link = braintrust.permalink(span.export())
    assert link == "https://braintrust.dev/noop-span"



"""
This module implements a two-layer caching system for Braintrust prompts.

The caching system consists of:
1. A fast in-memory LRU cache for frequently accessed prompts
2. A persistent disk-based cache that serves as a backing store

This allows for efficient prompt retrieval while maintaining persistence across sessions.
The cache is keyed by project identifier (ID or name), prompt slug, and version.
"""

from typing import Optional

from braintrust import prompt
from braintrust.prompt_cache import disk_cache, lru_cache


def _create_cache_key(project_id: Optional[str], project_name: Optional[str], slug: str, version: str) -> str:
    """Creates a unique cache key from project identifier, slug and version."""
    prefix = project_id or project_name
    if not prefix:
        raise ValueError("Either project_id or project_name must be provided")
    return f"{prefix}:{slug}:{version}"


class PromptCache:
    """
    A two-layer cache for Braintrust prompts with both in-memory and filesystem storage.

    This cache implements either a one or two-layer caching strategy:
    1. A fast in-memory LRU cache for frequently accessed prompts.
    2. An optional persistent filesystem-based cache that serves as a backing store.
    """

    def __init__(
        self,
        memory_cache: lru_cache.LRUCache[str, prompt.PromptSchema],
        disk_cache: Optional[disk_cache.DiskCache[prompt.PromptSchema]] = None,
    ):
        """
        Initialize the prompt cache.

        Args:
            memory_cache: The memory cache to use.
            disk_cache: Optional disk cache to use as backing store.
        """
        self.memory_cache = memory_cache
        self.disk_cache = disk_cache

    def get(
        self, slug: str, version: str, project_id: Optional[str] = None, project_name: Optional[str] = None
    ) -> prompt.PromptSchema:
        """
        Retrieve a prompt from the cache.

        Args:
            slug: The unique identifier for the prompt within its project.
            version: The version of the prompt.
            project_id: The ID of the project containing the prompt.
            project_name: The name of the project containing the prompt.

        Returns:
            The cached Prompt object.

        Raises:
            ValueError: If neither project_id nor project_name is provided.
            KeyError: If the prompt is not found in the cache.
        """
        cache_key = _create_cache_key(project_id, project_name, slug, version)

        # First check memory cache.
        try:
            return self.memory_cache.get(cache_key)
        except KeyError:
            pass

        # If not in memory and disk cache exists, check disk cache.
        if self.disk_cache:
            prompt = self.disk_cache.get(cache_key)
            if prompt is None:
                raise KeyError(f"Prompt not found in cache: {cache_key}")

            # Store in memory cache.
            self.memory_cache.set(cache_key, prompt)
            return prompt

        raise KeyError(f"Prompt not found in cache: {cache_key}")

    def set(
        self,
        slug: str,
        version: str,
        value: prompt.PromptSchema,
        project_id: Optional[str] = None,
        project_name: Optional[str] = None,
    ) -> None:
        """
        Store a prompt in the cache.

        Args:
            slug: The unique identifier for the prompt within its project.
            version: The version of the prompt.
            value: The Prompt object to store.
            project_id: The ID of the project containing the prompt.
            project_name: The name of the project containing the prompt.

        Raises:
            ValueError: If neither project_id nor project_name is provided.
            RuntimeError: If there is an error writing to the disk cache.
        """
        cache_key = _create_cache_key(project_id, project_name, slug, version)

        # Update memory cache.
        self.memory_cache.set(cache_key, value)

        # Update disk cache if available.
        if self.disk_cache:
            self.disk_cache.set(cache_key, value)



import logging
import time
from concurrent.futures import ThreadPoolExecutor

# pylint: disable=no-name-in-module
from ...aws import LazyClient, cloudformation

_logger = logging.getLogger("braintrust.install.logs")


def build_parser(subparsers, parents):
    parser = subparsers.add_parser(
        "bump-versions", help="Bump the version of each serverless function", parents=parents
    )
    parser.add_argument("name", help="Name of the CloudFormation stack")
    parser.set_defaults(func=main)


def main(args):
    resources = cloudformation.describe_stack_resources(StackName=args.name)
    lambda_ = LazyClient("lambda")

    # Then, publish the API handler lambdas just to bump their environment variable values
    api_handler = [r for r in resources["StackResources"] if r["LogicalResourceId"] == "APIHandler"]
    api_handler_js = [r for r in resources["StackResources"] if r["LogicalResourceId"] == "APIHandlerJS"]
    proxy_handler = [r for r in resources["StackResources"] if r["LogicalResourceId"] == "AIProxyFn"]

    if not api_handler or not api_handler_js or not proxy_handler:
        raise ValueError("No APIHandler, APIHandlerJS, or AIProxyFn found in the stack.")

    api_handler = api_handler[0]
    api_handler_js = api_handler_js[0]
    proxy_handler = proxy_handler[0]

    # Publish a new version of the API handler and re-point the "live2" alias to it
    for resource, alias in [(api_handler, "live2"), (api_handler_js, "live"), (proxy_handler, "live")]:
        new_version = lambda_.publish_version(FunctionName=resource["PhysicalResourceId"])
        lambda_.update_alias(
            FunctionName=resource["PhysicalResourceId"],
            Name=alias,
            FunctionVersion=new_version["Version"],
        )



import os
import shutil
import tempfile
import time
import unittest
from typing import Any

from braintrust import prompt
from braintrust.prompt_cache import disk_cache


class TestDiskCache(unittest.TestCase):
    def setUp(self):
        self.cache_dir = tempfile.mkdtemp()
        self.cache = disk_cache.DiskCache[dict](cache_dir=self.cache_dir, max_size=3)

    def tearDown(self):
        try:
            shutil.rmtree(self.cache_dir, ignore_errors=True)
        except Exception:
            pass

    def test_store_and_retrieve_values(self):
        test_data = {"foo": "bar"}
        self.cache.set("test-key", test_data)
        result = self.cache.get("test-key")
        self.assertEqual(result, test_data)

    def test_raise_keyerror_for_missing_keys(self):
        """Test raising KeyError for missing keys."""
        with self.assertRaises(KeyError) as cm:
            self.cache.get("missing-key")
        self.assertEqual(str(cm.exception), "'Cache key not found: missing-key'")

    def test_raise_keyerror_after_eviction(self):
        """Test that accessing evicted entries raises KeyError."""
        # Fill cache beyond max size (3).
        for i in range(3):
            self.cache.set(f"key{i}", {"value": i})
            time.sleep(0.1)  # wait to ensure different mtimes

        # Add one more to trigger eviction.
        self.cache.set("key3", {"value": 3})

        # The oldest entry should raise KeyError.
        with self.assertRaises(KeyError) as cm:
            self.cache.get("key0")
        self.assertEqual(str(cm.exception), "'Cache key not found: key0'")

    def test_evict_oldest_entries_when_cache_is_full(self):
        # Fill cache beyond max size (3).
        for i in range(3):
            self.cache.set(f"key{i}", {"value": i})
            time.sleep(0.1)  # wait to ensure different mtimes

        # Add one more to trigger eviction.
        self.cache.set("key3", {"value": 3})

        # The oldest entry should be evicted.
        with self.assertRaises(KeyError) as cm:
            self.cache.get("key0")
        self.assertEqual(str(cm.exception), "'Cache key not found: key0'")

        # Newer entries should still exist.
        newer = self.cache.get("key2")
        self.assertEqual(newer, {"value": 2})

    def test_throw_when_write_fails(self):
        # Make cache directory read-only.
        os.makedirs(self.cache_dir, exist_ok=True)
        os.chmod(self.cache_dir, 0o444)

        # Should raise when write fails.
        with self.assertRaises(RuntimeError):
            self.cache.set("test", {"foo": "bar"})

    def test_throw_when_read_fails(self):
        self.cache.set("test-key", {"foo": "bar"})

        # Make cache directory unreadable.
        os.chmod(self.cache_dir, 0o000)

        # Should raise when trying to read.
        with self.assertRaises(RuntimeError):
            self.cache.get("test-key")

        # Restore permissions so cleanup can happen.
        os.chmod(self.cache_dir, 0o777)

    def test_throw_on_corrupted_data(self):
        self.cache.set("test-key", {"foo": "bar"})

        # Corrupt the file.
        file_path = os.path.join(self.cache_dir, "test-key")
        with open(file_path, "w") as f:
            f.write("invalid data")

        # Should raise on corrupted data.
        with self.assertRaises(RuntimeError):
            self.cache.get("test-key")

    def test_evict_oldest_throws_on_stat_error(self):
        """Test that eviction throws when it can't get mtime."""
        # Create some test entries.
        for i in range(3):
            self.cache.set(f"key{i}", {"value": i})

        # Make cache directory unreadable.
        os.chmod(self.cache_dir, 0o000)

        # Should raise when trying to get mtime.
        with self.assertRaises(RuntimeError):
            self.cache.set("key3", {"value": 3})

        # Restore permissions so cleanup can happen.
        os.chmod(self.cache_dir, 0o777)

    def test_evict_oldest_throws_on_unlink_error(self):
        """Test that eviction throws when it can't remove files."""
        # Create some test entries.
        for i in range(3):
            self.cache.set(f"key{i}", {"value": i})
            time.sleep(0.1)  # ensure different mtimes

        # Make cache directory read-only.
        os.chmod(self.cache_dir, 0o444)

        # Should raise when trying to remove oldest entry.
        with self.assertRaises(RuntimeError):
            self.cache.set("key3", {"value": 3})

    def test_store_and_retrieve_with_serialization(self):
        """Test storing and retrieving objects using custom serialization."""
        cache = disk_cache.DiskCache[prompt.PromptSchema](
            cache_dir=self.cache_dir,
            max_size=3,
            serializer=lambda x: x.as_dict(),
            deserializer=prompt.PromptSchema.from_dict_deep,
        )

        # Create a test prompt.
        test_prompt = prompt.PromptSchema(
            id="456",
            project_id="123",
            _xact_id="789",
            name="test-prompt",
            slug="test-prompt",
            description=None,
            prompt_data=prompt.PromptData(),
            tags=None,
        )

        # Store and retrieve.
        cache.set("test-key", test_prompt)
        result = cache.get("test-key")

        # Should get back a PromptSchema instance.
        self.assertIsInstance(result, prompt.PromptSchema)
        self.assertEqual(result.as_dict(), test_prompt.as_dict())

    def test_serializer_handles_complex_objects(self):
        """Test that serializer is used for complex nested objects."""
        cache = disk_cache.DiskCache[prompt.PromptSchema](
            cache_dir=self.cache_dir, serializer=lambda x: x.as_dict(), deserializer=prompt.PromptSchema.from_dict_deep
        )

        # Create a prompt with nested data.
        test_prompt = prompt.PromptSchema(
            id="456",
            project_id="123",
            _xact_id="789",
            name="test-prompt",
            slug="test-prompt",
            description="test description",
            prompt_data=prompt.PromptData(
                prompt=prompt.PromptCompletionBlock(
                    content="test",
                )
            ),
            tags=["tag1", "tag2"],
        )

        # Store and retrieve.
        cache.set("test-key", test_prompt)
        result = cache.get("test-key")

        self.assertEqual(result.as_dict(), test_prompt.as_dict())

    def test_throw_on_deserializer_error(self):
        """Test that deserializer errors are propagated."""

        def bad_deserializer(data: Any) -> prompt.PromptSchema:
            raise ValueError("Deserialization failed")

        cache = disk_cache.DiskCache[prompt.PromptSchema](
            cache_dir=self.cache_dir, serializer=lambda x: x.as_dict(), deserializer=bad_deserializer
        )

        # Store a prompt.
        test_prompt = prompt.PromptSchema(
            id="456",
            project_id="123",
            _xact_id="789",
            name="test-prompt",
            slug="test-prompt",
            description=None,
            prompt_data=prompt.PromptData(),
            tags=None,
        )
        cache.set("test-key", test_prompt)

        # Should raise when deserializer fails.
        with self.assertRaises(RuntimeError) as cm:
            cache.get("test-key")
        self.assertIn("Deserialization failed", str(cm.exception))


if __name__ == "__main__":
    unittest.main()



"""Re-exports generated typespecs that are considered public API."""

from ._types import (
    AttachmentReference,
    AttachmentStatus,
    ChatCompletionMessageParam,
    DatasetEvent,
    ExperimentEvent,
    IfExists,
    ModelParams,
    PromptData,
    PromptOptions,
    SavedFunctionId,
    SpanAttributes,
    SpanType,
    ToolFunctionDefinition,
)



import unittest
from typing import List

from .util import LazyValue


class TestLazyValue(unittest.TestCase):
    def test_evaluates_exactly_once(self):
        call_count = 0

        def compute_value():
            nonlocal call_count
            call_count += 1
            return "test"

        lazy = LazyValue(compute_value, use_mutex=True)

        self.assertEqual(call_count, 0)
        self.assertFalse(lazy.has_succeeded)

        # First access should compute
        value1 = lazy.get()
        self.assertEqual(value1, "test")
        self.assertEqual(call_count, 1)
        self.assertTrue(lazy.has_succeeded)

        # Second access should use cached value
        value2 = lazy.get()
        self.assertEqual(value2, "test")
        self.assertEqual(call_count, 1)
        self.assertTrue(lazy.has_succeeded)

    def test_has_succeeded_only_set_after_success(self):
        def failing_compute():
            raise ValueError("test error")

        lazy = LazyValue(failing_compute, use_mutex=True)

        self.assertFalse(lazy.has_succeeded)
        self.assertIsNone(lazy.value)

        with self.assertRaises(ValueError) as ctx:
            lazy.get()

        self.assertEqual(str(ctx.exception), "test error")
        self.assertFalse(lazy.has_succeeded)
        self.assertIsNone(lazy.value)

    def test_thread_safety(self):
        import threading
        import time

        # This will be used to track if multiple threads try to compute simultaneously
        computing = threading.Event()

        def compute_value():
            # If computing is already set when we enter, another thread
            # is also trying to compute - this should never happen
            if computing.is_set():
                raise RuntimeError("Concurrent computation detected!")

            computing.set()
            try:
                # Sleep briefly to increase chance of race conditions
                time.sleep(0.1)
                return "test result"
            finally:
                computing.clear()

        lazy = LazyValue(compute_value, use_mutex=True)

        # Launch multiple threads that all try to get() simultaneously
        threads: List[threading.Thread] = []
        results: List[str] = []
        errors: List[Exception] = []

        def worker():
            try:
                results.append(lazy.get())
            except Exception as e:
                errors.append(e)

        for _ in range(10):
            t = threading.Thread(target=worker)
            threads.append(t)
            t.start()

        # Wait for all threads to complete
        for t in threads:
            t.join()

        # Verify no errors occurred
        self.assertEqual(errors, [])

        # Verify all threads got the same result
        self.assertEqual(len(results), 10)
        self.assertTrue(all(r == "test result" for r in results))

        # Verify the computation is marked succeeded.
        self.assertTrue(lazy.has_succeeded)


if __name__ == "__main__":
    unittest.main()



"""
This module provides classes and functions for handling Braintrust streams.

A Braintrust stream is a wrapper around a generator of `BraintrustStreamChunk`,
with utility methods to make them easy to log and convert into various formats.
"""

import dataclasses
import json
from itertools import tee
from typing import Generator, List, Literal, Union

from sseclient import SSEClient


@dataclasses.dataclass
class BraintrustTextChunk:
    """
    A chunk of text data from a Braintrust stream.
    """

    data: str
    type: Literal["text_delta"] = "text_delta"


@dataclasses.dataclass
class BraintrustJsonChunk:
    """
    A chunk of JSON data from a Braintrust stream.
    """

    data: str
    type: Literal["json_delta"] = "json_delta"


@dataclasses.dataclass
class BraintrustErrorChunk:
    """
    An error chunk from a Braintrust stream.
    """

    data: str
    type: Literal["error"] = "error"


@dataclasses.dataclass
class BraintrustConsoleChunk:
    """
    A console chunk from a Braintrust stream.
    """

    message: str
    stream: Literal["stderr", "stdout"]
    type: Literal["console"] = "console"


@dataclasses.dataclass
class BraintrustProgressChunk:
    """
    A progress chunk from a Braintrust stream.
    """

    data: str
    id: str
    object_type: str
    format: str
    output_type: str
    name: str
    event: Literal["json_delta", "text_delta"]
    type: Literal["progress"] = "progress"


class BraintrustInvokeError(ValueError):
    """
    An error that occurs during a Braintrust stream.
    """

    pass


BraintrustStreamChunk = Union[BraintrustTextChunk, BraintrustJsonChunk, BraintrustErrorChunk]


class BraintrustStream:
    """
    A Braintrust stream. This is a wrapper around a generator of `BraintrustStreamChunk`,
    with utility methods to make them easy to log and convert into various formats.
    """

    def __init__(self, base_stream: Union[SSEClient, List[BraintrustStreamChunk]]):
        """
        Initialize a BraintrustStream.

        Args:
            base_stream: Either an SSEClient or a list of BraintrustStreamChunks.
        """
        if isinstance(base_stream, SSEClient):
            self.stream = self._parse_sse_stream(base_stream)
        else:
            self.stream = base_stream
        self._memoized_final_value = None

    def _parse_sse_stream(self, sse_client: SSEClient) -> Generator[BraintrustStreamChunk, None, None]:
        """
        Parse an SSE stream into BraintrustStreamChunks.

        Args:
            sse_client: The SSEClient to parse.

        Yields:
            BraintrustStreamChunk: Parsed chunks from the SSE stream.
        """
        for event in sse_client.events():
            if event.event == "text_delta":
                yield BraintrustTextChunk(data=json.loads(event.data))
            elif event.event == "json_delta":
                yield BraintrustJsonChunk(data=event.data)
            elif event.event == "error":
                yield BraintrustErrorChunk(data=json.loads(event.data))
            elif event.event == "console":
                event_data = json.loads(event.data)
                yield BraintrustConsoleChunk(
                    message=event_data["message"],
                    stream=event_data["stream"],
                )
            elif event.event == "progress":
                event_data = json.loads(event.data)
                yield BraintrustProgressChunk(
                    data=event_data["data"],
                    id=event_data["id"],
                    object_type=event_data["object_type"],
                    format=event_data["format"],
                    output_type=event_data["output_type"],
                    name=event_data["name"],
                    event=event_data["event"],
                )

    def copy(self):
        """
        Copy the stream. This returns a new stream that shares the same underlying
        generator (via `tee`). Since generators are consumed in Python, use `copy()` if you
        need to use the stream multiple times.

        Returns:
            BraintrustStream: A new stream that you can independently consume.
        """
        current_stream = self.stream
        self.stream, new_stream = tee(current_stream)
        return BraintrustStream(new_stream)

    def final_value(self):
        """
        Get the final value of the stream. The final value is the concatenation of all
        the chunks in the stream, deserialized into a string or object, depending on
        the value's type.

        This function consumes the stream, so if you need to use the stream multiple
        times, you should call `copy()` first.

        Returns:
            The final value of the stream.
        """
        if self._memoized_final_value is None:
            self._memoized_final_value = parse_stream(self)
        return self._memoized_final_value

    def __iter__(self):
        """
        Iterate over the stream chunks.

        Yields:
            BraintrustStreamChunk: The next chunk in the stream.
        """
        yield from self.stream


def parse_stream(stream: BraintrustStream):
    """
    Parse a BraintrustStream into its final value.

    Args:
        stream: The BraintrustStream to parse.

    Returns:
        The final value of the stream.
    """
    text_chunks = []
    json_chunks = []

    for chunk in stream:
        if isinstance(chunk, BraintrustTextChunk):
            text_chunks.append(chunk.data)
        elif isinstance(chunk, BraintrustJsonChunk):
            json_chunks.append(chunk.data)
        elif isinstance(chunk, BraintrustErrorChunk):
            raise BraintrustInvokeError(chunk.data)
        elif isinstance(chunk, BraintrustProgressChunk) or isinstance(chunk, BraintrustConsoleChunk):
            pass
        else:
            raise ValueError(f"Unknown chunk type (you may need to update the SDK): {type(chunk)}")

    if json_chunks:
        return json.loads("".join(json_chunks))
    elif text_chunks:
        return "".join(text_chunks)
    else:
        return None



import os
import shutil
import tempfile
import unittest

from braintrust import prompt
from braintrust.prompt_cache import disk_cache, lru_cache, prompt_cache


class TestPromptCache(unittest.TestCase):
    def setUp(self):
        # Create a temporary directory for each test
        self.cache_dir = tempfile.mkdtemp()
        mc = lru_cache.LRUCache[str, prompt.PromptSchema](max_size=2)
        dc = disk_cache.DiskCache[prompt.PromptSchema](
            cache_dir=self.cache_dir,
            max_size=5,
            serializer=lambda x: x.as_dict(),
            deserializer=prompt.PromptSchema.from_dict_deep,
        )
        self.cache = prompt_cache.PromptCache(memory_cache=mc, disk_cache=dc)

        self.test_prompt = prompt.PromptSchema(
            id="456",
            project_id="123",
            _xact_id="789",
            name="test-prompt",
            slug="test-prompt",
            description=None,
            prompt_data=prompt.PromptData(),
            tags=None,
        )

    def tearDown(self):
        # Clean up the temporary directory.
        try:
            shutil.rmtree(self.cache_dir, ignore_errors=True)
        except Exception:
            pass

    def test_store_and_retrieve_from_memory_cache(self):
        self.cache.set("test-prompt", "789", self.test_prompt, project_id="123")
        result = self.cache.get("test-prompt", version="789", project_id="123")
        self.assertEqual(result.as_dict(), self.test_prompt.as_dict())

    def test_work_with_project_name(self):
        self.cache.set("test-prompt", "789", self.test_prompt, project_name="test-project")
        result = self.cache.get("test-prompt", version="789", project_name="test-project")
        self.assertEqual(result.as_dict(), self.test_prompt.as_dict())

    def test_throw_error_if_no_project_identifier(self):
        with self.assertRaisesRegex(ValueError, "Either project_id or project_name must be provided"):
            self.cache.get("test-prompt", version="789")

        with self.assertRaisesRegex(ValueError, "Either project_id or project_name must be provided"):
            self.cache.set("test-prompt", "789", self.test_prompt)

    def test_store_and_retrieve_from_disk_after_memory_eviction(self):
        # Fill memory cache (max size is 2).
        self.cache.set("test-prompt", "789", self.test_prompt, project_id="123")
        self.cache.set("prompt2", "789", self.test_prompt, project_id="123")
        self.cache.set("prompt3", "789", self.test_prompt, project_id="123")

        # Original prompt should now be on disk but not in memory.
        result = self.cache.get("test-prompt", version="789", project_id="123")
        self.assertEqual(result.as_dict(), self.test_prompt.as_dict())

    def test_raise_for_nonexistent_prompts(self):
        with self.assertRaises(KeyError):
            self.cache.get("missing-prompt", version="789", project_id="123")

    def test_handle_different_projects_with_same_slug(self):
        self.cache.set("test-prompt", "789", self.test_prompt, project_id="123")

        different_prompt = prompt.PromptSchema.from_dict_deep(self.test_prompt.as_dict())
        different_prompt.project_id = "different-project"
        self.cache.set("test-prompt", "789", different_prompt, project_id="different-project")

        result1 = self.cache.get("test-prompt", version="789", project_id="123")
        result2 = self.cache.get("test-prompt", version="789", project_id="different-project")

        self.assertEqual(result1.project_id, "123")
        self.assertEqual(result2.project_id, "different-project")

    def test_memory_only_cache(self):
        memory_only_cache = prompt_cache.PromptCache(memory_cache=lru_cache.LRUCache(max_size=2))

        memory_only_cache.set("test-prompt", "789", self.test_prompt, project_id="123")
        result = memory_only_cache.get("test-prompt", version="789", project_id="123")
        self.assertEqual(result.as_dict(), self.test_prompt.as_dict())

        # Fill memory cache beyond capacity.
        memory_only_cache.set("prompt2", "789", self.test_prompt, project_id="123")
        memory_only_cache.set("prompt3", "789", self.test_prompt, project_id="123")

        # First prompt should be gone since there's no disk backup.
        with self.assertRaises(KeyError):
            memory_only_cache.get("test-prompt", version="789", project_id="123")

    def test_throw_when_disk_write_fails(self):
        # Make cache directory read-only.
        os.chmod(self.cache_dir, 0o444)

        # Should raise when disk write fails.
        with self.assertRaises(RuntimeError):
            self.cache.set("test-prompt", "789", self.test_prompt, project_id="123")

        # Memory cache should still be updated despite disk failure.
        result = self.cache.get("test-prompt", version="789", project_id="123")
        self.assertEqual(result.as_dict(), self.test_prompt.as_dict())

        # Restore permissions so cleanup can happen.
        os.chmod(self.cache_dir, 0o777)


if __name__ == "__main__":
    unittest.main()
